"""
Tests for the molecule module.
"""
import pytest
import numpy as np
import gimbal.molecule as molecule
from examples import Geometries as eg
from examples import Files as ef


def test_empty_BaseMolecule():
    mol = molecule.BaseMolecule()
    assert len(mol.elem) == 0
    assert len(mol.xyz) == 0
    assert len(mol.vec) == 0
    assert mol.comment == ''
    assert str(mol) == '[]'
    assert repr(mol) == 'BaseMolecule(\'\', [])'


def test_single_atom_vec_BaseMolecule():
    mol = molecule.BaseMolecule(*eg.he, vec=np.ones(3))
    zf = 3*'{:14.8f}'.format(0)
    of = 3*'{:14.8f}'.format(1)
    ze = ','.join(3*['{:16.8e}'.format(0)])
    oe = ','.join(3*['{:16.8e}'.format(1)])
    assert mol.natm == 1
    assert np.all(mol.elem == ['He'])
    assert np.allclose(mol.xyz, np.zeros((1, 3)))
    assert str(mol) == '[[He' + zf + '\n    ' + of + ']]'
    assert repr(mol) == ('BaseMolecule(\'\',\n [[He,' + ze +
                         ',\n      ' + oe + ']])')


def test_two_atom_comment_BaseMolecule():
    mol = molecule.BaseMolecule(*eg.h2, comment='comment')
    assert mol.natm == 2
    assert np.all(mol.elem == eg.h2[0])
    assert np.allclose(mol.xyz, eg.h2[1])
    assert 'comment' in str(mol)


def test_eleven_atom_BaseMolecule():
    mol = molecule.BaseMolecule(*eg.c3h8)
    assert mol.natm == 11
    assert np.all(mol.elem == eg.c3h8[0])
    assert np.allclose(mol.xyz, eg.c3h8[1])
    assert '...' in str(mol)
    assert '...' in repr(mol)


def test_BaseMolecule_xyz_not_3d():
    with pytest.raises(ValueError, match=r'Molecular geometry must be 3.*'):
        mol = molecule.BaseMolecule('H', np.zeros(2))


def test_BaseMolecule_vec_not_3d():
    with pytest.raises(ValueError, match=r'Molecular vector must be 3.*'):
        mol = molecule.BaseMolecule('H', np.zeros(3), vec=np.ones(2))


def test_BaseMolecule_len_elem_not_equal_len_xyz():
    with pytest.raises(ValueError, match=r'Number of element labels .*'):
        mol = molecule.BaseMolecule('H', np.zeros((2, 3)))


def test_BaseMolecule_xyz_vec_different_shape():
    with pytest.raises(ValueError, match=r'Cartesian geometry and vector .*'):
        mol = molecule.BaseMolecule(['H', 'H'], np.zeros((2, 3)),
                                    np.ones((1, 3)))


def test_BaseMolecule_copy():
    mol1 = molecule.BaseMolecule(*eg.c2h4)
    mol2 = mol1.copy()
    assert np.all(mol1.elem == eg.c2h4[0])
    assert np.all(mol2.elem == eg.c2h4[0])
    assert np.allclose(mol1.xyz, eg.c2h4[1])
    assert np.allclose(mol2.xyz, eg.c2h4[1])


def test_BaseMolecule_save():
    mol = molecule.BaseMolecule(*eg.he)
    mol.elem = np.array(['H'])
    mol.xyz = np.ones((1, 3))
    mol.comment = 'comment'
    mol.save()
    assert mol.saved
    assert np.all(mol.save_elem == ['H'])
    assert np.allclose(mol.save_xyz, np.ones((1, 3)))
    assert mol.save_comment == 'comment'


def test_BaseMolecule_revert():
    mol = molecule.BaseMolecule(*eg.he)
    mol.elem = np.array(['H'])
    mol.xyz = np.ones((1, 3))
    mol.comment = 'comment'
    mol.revert()
    assert mol.saved
    assert np.all(mol.elem == eg.he[0])
    assert np.allclose(mol.xyz, eg.he[1])
    assert mol.comment == ''


def test_BaseMolecule_add_atoms_single():
    mol = molecule.BaseMolecule(*eg.he)
    mol.add_atoms('H', np.ones(3))
    soln_elem = np.hstack((eg.he[0], 'H'))
    soln_xyz = np.vstack((eg.he[1], np.ones(3)))
    assert np.all(mol.elem == soln_elem)
    assert np.allclose(mol.xyz, soln_xyz)
    assert not mol.saved


def test_BaseMolecule_add_atoms_multiple():
    mol = molecule.BaseMolecule(*eg.he)
    mol.add_atoms(*eg.ch4)
    soln_elem = np.hstack((eg.he[0], eg.ch4[0]))
    soln_xyz = np.vstack((eg.he[1], eg.ch4[1]))
    assert np.all(mol.elem == soln_elem)
    assert np.allclose(mol.xyz, soln_xyz)
    assert not mol.saved


def test_BaseMolecule_add_atoms_new_vec():
    mol = molecule.BaseMolecule(*eg.he)
    mol.add_atoms('H', np.ones(3), new_vec=2*np.ones(3))
    soln_elem = np.hstack((eg.he[0], 'H'))
    soln_xyz = np.vstack((eg.he[1], np.ones(3)))
    soln_vec = np.vstack((np.zeros(3), 2*np.ones(3)))
    assert np.all(mol.elem == soln_elem)
    assert np.allclose(mol.xyz, soln_xyz)
    assert np.allclose(mol.vec, soln_vec)


def test_BaseMolecule_rm_atoms_single():
    mol = molecule.BaseMolecule(*eg.ch4)
    mol.rm_atoms(0)
    assert np.all(mol.elem == eg.ch4[0][1:])
    assert np.allclose(mol.xyz, eg.ch4[1][1:])
    assert not mol.saved


def test_BaseMolecule_rm_atoms_multiple():
    mol = molecule.BaseMolecule(*eg.ch4)
    mol.rm_atoms((1, 2, 3, 4))
    assert np.all(mol.elem == eg.ch4[0][0])
    assert np.allclose(mol.xyz, eg.ch4[1][0])
    assert not mol.saved


def test_BaseMolecule_rearrange_all():
    mol = molecule.BaseMolecule(*eg.ch4)
    ind = [4, 2, 3, 0, 1]
    mol.rearrange(ind)
    assert np.all(mol.elem == eg.ch4[0][ind])
    assert np.allclose(mol.xyz, eg.ch4[1][ind])
    assert not mol.saved


def test_BaseMolecule_rearrange_old_ind():
    mol = molecule.BaseMolecule(*eg.ch4)
    ninds = [1, 2]
    oinds = [3, 4]
    mol.rearrange(ninds, old_ind=oinds)
    soln_elem = eg.ch4[0][[0, 3, 4, 1, 2]]
    soln_xyz = eg.ch4[1][[0, 3, 4, 1, 2]]
    assert np.all(mol.elem == soln_elem)
    assert np.allclose(mol.xyz, soln_xyz)
    assert not mol.saved


def test_BaseMolecule_rearrange_fails():
    mol = molecule.BaseMolecule(*eg.ch4)
    with pytest.raises(IndexError, match=r'Old and new indices must be .*'):
        mol.rearrange([1, 3])


def test_empty_Molecule():
    mol = molecule.Molecule()
    assert len(mol.elem) == 0
    assert len(mol.xyz) == 0
    assert len(mol.vec) == 0
    assert mol.comment == ''
    assert str(mol) == '[]'
    assert repr(mol) == 'Molecule(\'\', [])'


def test_single_atom_vec_Molecule():
    mol = molecule.Molecule(*eg.he, vec=np.ones(3))
    zf = 3*'{:14.8f}'.format(0)
    of = 3*'{:14.8f}'.format(1)
    ze = ','.join(3*['{:16.8e}'.format(0)])
    oe = ','.join(3*['{:16.8e}'.format(1)])
    assert mol.natm == 1
    assert np.all(mol.elem == ['XM', 'He'])
    assert np.allclose(mol.xyz, np.zeros((2, 3)))
    assert str(mol) == '[[He' + zf + '\n    ' + of + ']]'
    assert repr(mol) == ('Molecule(\'\',\n [[XM,' + ze + ',\n      ' + ze +
                         '],\n  [He,' + ze + ',\n      ' + oe + ']])')


def test_eleven_atom_Molecule():
    mol = molecule.Molecule(*eg.c3h8)
    assert mol.natm == 11
    assert np.all(mol.elem == np.hstack(('XM', eg.c3h8[0])))
    assert np.allclose(mol.xyz, np.vstack((np.zeros(3), eg.c3h8[1])), atol=1e-7)
    assert '...' in str(mol)
    assert '...' in repr(mol)


def test_Molecule_add_Molecule():
    mol1 = molecule.Molecule(*eg.he)
    mol2 = molecule.Molecule(*eg.c2h4)
    bund = mol1 + mol2
    assert np.all(bund.molecules[0].elem[1:] == eg.he[0])
    assert np.all(bund.molecules[0].xyz[1:] == eg.he[1])
    assert np.all(bund.molecules[1].elem[1:] == eg.c2h4[0])
    assert np.all(bund.molecules[1].xyz[1:] == eg.c2h4[1])


def test_Molecule_add_MoleculeBundle():
    mol = molecule.Molecule(*eg.he)
    bund1 = molecule.MoleculeBundle(molecule.Molecule(*eg.c2h4))
    bund = mol + bund1
    assert np.all(bund.molecules[0].elem[1:] == eg.he[0])
    assert np.all(bund.molecules[0].xyz[1:] == eg.he[1])
    assert np.all(bund.molecules[1].elem[1:] == eg.c2h4[0])
    assert np.all(bund.molecules[1].xyz[1:] == eg.c2h4[1])


def test_Molecule_copy():
    mol1 = molecule.Molecule(*eg.c2h4)
    mol2 = mol1.copy()
    assert np.all(mol1.elem[1:] == eg.c2h4[0])
    assert np.all(mol2.elem[1:] == eg.c2h4[0])
    assert np.allclose(mol1.xyz[1:], eg.c2h4[1])
    assert np.allclose(mol2.xyz[1:], eg.c2h4[1])


def test_Molecule_rearrange_all():
    mol = molecule.Molecule(*eg.ch4)
    ind = [5, 3, 4, 1, 2]
    oind = [i-1 for i in ind]
    mol.rearrange(ind)
    assert np.all(mol.elem[1:] == eg.ch4[0][oind])
    assert np.allclose(mol.xyz[1:], eg.ch4[1][oind])
    assert not mol.saved


def test_Molecule_read_filename(tmpdir):
    f = tmpdir.join('tmp.xyz')
    f.write(ef.xyz_novec)
    mol = molecule.Molecule()
    mol.read(str(f.realpath()))
    assert np.all(mol.elem[1:] == eg.ch4[0])
    assert np.allclose(mol.xyz[1:], eg.ch4[1])
    assert not mol.saved


def test_Molecule_read_openfile_vec(tmpdir):
    f = ef.tmpf(tmpdir, ef.xyz_vec)
    mol = molecule.Molecule()
    mol.read(f, hasvec=True)
    assert np.all(mol.elem[1:] == eg.ch4[0])
    assert np.allclose(mol.xyz[1:], eg.ch4[1])
    assert not mol.saved


def test_Molecule_write_filename(tmpdir):
    f = tmpdir.join('tmp.xyz')
    mol = molecule.Molecule(*eg.ch4, comment='comment line')
    mol.write(str(f.realpath()))
    assert f.read() == ef.xyz_novec


def test_Molecule_write_openfile(tmpdir):
    f = tmpdir.join('tmp.xyz')
    mol = molecule.Molecule(*eg.ch4, comment='comment line')
    mol.write(f.open(mode='w'))
    assert f.read() == ef.xyz_novec


def test_Molecule_get_mass():
    mol = molecule.Molecule(*eg.ch4)
    mass = mol.get_mass()
    soln = [0.00000000] + [12.00000000] + 4*[1.00782504]
    assert np.allclose(mass, soln)


def test_Molecule_get_formula_single_atom():
    mol = molecule.Molecule(*eg.he)
    form = mol.get_formula()
    assert form == 'He'


def test_Molecule_get_formula_multiple_atoms():
    mol = molecule.Molecule(*eg.c2h4)
    form = mol.get_formula()
    assert form == 'C2H4'


def test_Molecule_measure():
    mol = molecule.Molecule(*eg.c2h4_ms)
    oop = mol.measure('oop', 2, 4, 3, 1)
    assert np.isclose(oop, -np.pi/4)


def test_Molecule_centre_mass():
    mol = molecule.Molecule(*eg.ch4)
    mol.xyz[1] += 2
    mol.centre_mass()
    mass = mol.get_mass()
    cm = np.sum(mass[:,np.newaxis] * mol.xyz, axis=0) / np.sum(mass)
    assert np.allclose(mol.xyz[0], np.zeros(3))
    assert np.allclose(cm, np.zeros(3))


def test_Molecule_translate():
    mol = molecule.Molecule(*eg.he)
    mol.translate(1, [1, 1, 1])
    assert np.allclose(mol.xyz, np.ones(3)/np.sqrt(3))


def test_Molecule_rotate_no_vec():
    mol = molecule.Molecule(*eg.c2h4)
    mol.rotate(np.pi/2, 'Z')
    soln = np.array([-eg.c2h4[1][:,1], -eg.c2h4[1][:,0], eg.c2h4[1][:,2]]).T
    assert np.allclose(mol.xyz[1:], soln)


def test_Molecule_rotate_vec():
    mol = molecule.Molecule(*eg.c2h4, vec=np.ones((6, 3)))
    mol.rotate(np.pi/2, 'Z')
    soln = np.ones_like(eg.c2h4[1])
    soln[:,0] = -1
    assert np.allclose(mol.vec[1:], soln)


def test_Molecule_match_to_ref():
    mol5 = molecule.Molecule(eg.ch2f2[0], eg.ch2f2[5])
    mol1 = molecule.Molecule(eg.ch2f2[0], eg.ch2f2[1])
    mol2 = molecule.Molecule(eg.ch2f2[0], eg.ch2f2[2])
    bund = molecule.MoleculeBundle([mol1, mol2])
    ind = mol5.match_to_ref(bund)
    assert np.all(np.abs(mol5.xyz - mol2.xyz) < 1)
    assert not np.all(np.abs(mol5.xyz - mol1.xyz) < 1)
    assert ind == 1


def test_Molecule_subst():
    mol = molecule.Molecule(*eg.c2h4)
    mol.subst('ch3', 3, pl=2)
    assert np.all(mol.elem[1:] == eg.c2h3me[0])
    assert np.allclose(mol.xyz[1:], eg.c2h3me[1])


def test_Molecule_subst_vec():
    mol = molecule.Molecule(*eg.c2h4)
    mol.print_vec = True
    mol.vec = np.ones_like(mol.xyz)
    mol.subst('ch3', 3, pl=2)
    newvec = np.vstack((np.ones((2, 3)), np.zeros((4, 3)), np.ones((3, 3))))
    assert np.all(mol.elem[1:] == eg.c2h3me[0])
    assert np.allclose(mol.xyz[1:], eg.c2h3me[1])
    assert np.allclose(mol.vec[1:], newvec)


def test_empty_MoleculeBundle():
    bund = molecule.MoleculeBundle()
    assert len(bund.molecules) == 0
    assert bund.nmol == 0
    assert str(bund) == '[]'
    assert repr(bund) == 'MoleculeBundle()'


def test_single_atom_MoleculeBundle():
    mol = molecule.Molecule(*eg.he)
    bund = molecule.MoleculeBundle(mol)
    zf = 3*'{:14.8f}'.format(0)
    ze = 3*',{:16.8e}'.format(0)
    assert bund.nmol == 1
    assert str(bund) == '[[[He' + zf + ']]]'
    assert repr(bund) == ('MoleculeBundle(\n Molecule(\'\',\n  [[XM' + ze +
                          '],\n   [He' + ze + ']]))')


def test_seven_molecule_MoleculeBundle():
    mol = molecule.Molecule(*eg.he)
    bund = molecule.MoleculeBundle(7*[mol])
    assert bund.nmol == 7
    assert '...' in str(bund)
    assert '...' in repr(bund)


def test_MoleculeBundle_add_MoleculeBundle():
    mol1 = molecule.Molecule(*eg.he)
    mol2 = molecule.Molecule(*eg.ch4)
    mol3 = molecule.Molecule(*eg.c2h4)
    bund1 = molecule.MoleculeBundle(mol1)
    bund23 = molecule.MoleculeBundle([mol2, mol3])
    bund = bund1 + bund23
    assert np.all(bund.molecules[0].elem[1:] == eg.he[0])
    assert np.all(bund.molecules[0].xyz[1:] == eg.he[1])
    assert np.all(bund.molecules[1].elem[1:] == eg.ch4[0])
    assert np.all(bund.molecules[1].xyz[1:] == eg.ch4[1])
    assert np.all(bund.molecules[2].elem[1:] == eg.c2h4[0])
    assert np.all(bund.molecules[2].xyz[1:] == eg.c2h4[1])


def test_MoleculeBundle_add_wrong_type():
    bund1 = molecule.MoleculeBundle(molecule.Molecule(*eg.he))
    with pytest.raises(TypeError, match=r'Addition not supported for types .*'):
        bund = bund1 + np.ones(3)


def test_MoleculeBundle_iadd_Molecule():
    mol1 = molecule.Molecule(*eg.he)
    mol2 = molecule.Molecule(*eg.ch4)
    mol3 = molecule.Molecule(*eg.c2h4)
    bund = molecule.MoleculeBundle([mol1, mol2])
    bund += mol3
    assert np.all(bund.molecules[0].elem[1:] == eg.he[0])
    assert np.all(bund.molecules[0].xyz[1:] == eg.he[1])
    assert np.all(bund.molecules[1].elem[1:] == eg.ch4[0])
    assert np.all(bund.molecules[1].xyz[1:] == eg.ch4[1])
    assert np.all(bund.molecules[2].elem[1:] == eg.c2h4[0])
    assert np.all(bund.molecules[2].xyz[1:] == eg.c2h4[1])


def test_MoleculeBundle_wrong_type():
    with pytest.raises(TypeError, match=r'Elements of molecule bundle must .*'):
        bund = molecule.MoleculeBundle(eg.he)


def test_MoleculeBundle_copy():
    mol1 = molecule.Molecule(*eg.he)
    mol2 = molecule.Molecule(*eg.ch4)
    bund1 = molecule.MoleculeBundle([mol1, mol2])
    bund2 = bund1.copy()
    assert np.all(bund1.molecules[0].elem == bund2.molecules[0].elem)
    assert np.all(bund1.molecules[1].elem == bund2.molecules[1].elem)
    assert np.allclose(bund1.molecules[0].xyz, bund2.molecules[0].xyz)
    assert np.allclose(bund1.molecules[1].xyz, bund2.molecules[1].xyz)


def test_MoleculeBundle_save():
    mol = molecule.Molecule(*eg.he)
    bund = molecule.MoleculeBundle([mol, mol])
    bund.molecules = [mol]
    bund.molecules[0].xyz += 1
    bund.molecules[0].elem[1] = 'H'
    bund.save()
    assert bund.saved
    assert bund.molecules[0].saved
    assert len(bund.save_molecules) == 1


def test_MoleculeBundle_revert():
    mol = molecule.Molecule(*eg.he)
    bund = molecule.MoleculeBundle([mol, mol])
    bund.molecules = [mol]
    bund.molecules[0].xyz += 1
    bund.molecules[0].elem[1] = 'H'
    bund.revert()
    assert bund.saved
    assert bund.molecules[0].saved
    assert len(bund.save_molecules) == 2
    assert bund.molecules[1].elem[1] == 'He'


def test_MoleculeBundle_rearrange_all():
    mol1 = molecule.Molecule('He', np.zeros(3))
    mol2 = molecule.Molecule('Ne', np.ones(3))
    mol3 = molecule.Molecule('Ar', -np.ones(3))
    bund = molecule.MoleculeBundle([mol1, mol2, mol3])
    bund.rearrange([2, 0, 1])
    assert bund.molecules[0] == mol3
    assert bund.molecules[1] == mol1
    assert bund.molecules[2] == mol2
    assert not bund.saved


def test_MoleculeBundle_rearrange_old_ind():
    mol1 = molecule.Molecule('He', np.zeros(3))
    mol2 = molecule.Molecule('Ne', np.ones(3))
    mol3 = molecule.Molecule('Ar', -np.ones(3))
    bund = molecule.MoleculeBundle([mol1, mol2, mol3])
    bund.rearrange(1, old_ind=0)
    assert bund.molecules[0] == mol2
    assert bund.molecules[1] == mol1
    assert bund.molecules[2] == mol3
    assert not bund.saved


def test_MoleculeBundle_add_molecules():
    mol1 = molecule.Molecule('He', np.zeros(3))
    mol2 = molecule.Molecule('Ne', np.ones(3))
    mol3 = molecule.Molecule('Ar', -np.ones(3))
    bund = molecule.MoleculeBundle(mol1)
    bund.add_molecules([mol2, mol3])
    assert bund.molecules[0] == mol1
    assert bund.molecules[1] == mol2
    assert bund.molecules[2] == mol3
    assert not bund.saved


def test_MoleculeBundle_rm_molecules():
    mol1 = molecule.Molecule('He', np.zeros(3))
    mol2 = molecule.Molecule('Ne', np.ones(3))
    mol3 = molecule.Molecule('Ar', -np.ones(3))
    bund = molecule.MoleculeBundle([mol1, mol2, mol3])
    bund.rm_molecules(1)
    assert bund.nmol == 2
    assert bund.molecules[0] == mol1
    assert bund.molecules[1] == mol3
    assert not bund.saved


def test_MoleculeBundle_read_filename(tmpdir):
    f = tmpdir.join('tmp.xyz')
    f.write(ef.xyz_novec + ef.zmtvar_nocom)
    bund = molecule.MoleculeBundle()
    bund.read(str(f.realpath()))
    assert np.all(bund.molecules[0].elem[1:] == eg.ch4[0])
    assert np.all(bund.molecules[1].elem[1:] == eg.ch4[0])
    assert np.allclose(bund.molecules[0].xyz[1:], eg.ch4[1])
    assert np.allclose(bund.molecules[1].xyz[1:], eg.ch4_zmt[1])
    assert not bund.saved


def test_MoleculeBundle_read_openfile(tmpdir):
    f = ef.tmpf(tmpdir, ef.xyz_novec + ef.zmtvar_nocom)
    bund = molecule.MoleculeBundle()
    bund.read(f)
    assert np.all(bund.molecules[0].elem[1:] == eg.ch4[0])
    assert np.all(bund.molecules[1].elem[1:] == eg.ch4[0])
    assert np.allclose(bund.molecules[0].xyz[1:], eg.ch4[1])
    assert np.allclose(bund.molecules[1].xyz[1:], eg.ch4_zmt[1])
    assert not bund.saved


def test_MoleculeBundle_write_filename(tmpdir):
    f = tmpdir.join('tmp.xyz')
    mol = molecule.Molecule(*eg.ch4, comment='comment line')
    bund = molecule.MoleculeBundle([mol, mol])
    bund.write(str(f.realpath()))
    assert f.read() == 2*ef.xyz_novec


def test_MoleculeBundle_write_openfile(tmpdir):
    f = tmpdir.join('tmp.xyz')
    mol = molecule.Molecule(*eg.ch4, comment='comment line')
    bund = molecule.MoleculeBundle([mol, mol])
    bund.write(f.open(mode='w'))
    assert f.read() == 2*ef.xyz_novec


def test_MoleculeBundle_write_vec(tmpdir):
    f = tmpdir.join('tmp.xyz')
    vec = np.ones((5, 3))
    mol = molecule.Molecule(*eg.ch4, vec=vec, comment='comment line')
    bund = molecule.MoleculeBundle([mol, mol])
    bund.write(f.open(mode='w'))
    assert f.read() == 2*ef.xyz_vec


def test_MoleculeBundle_measure():
    mol1 = molecule.Molecule(eg.ch2f2[0], eg.ch2f2[1])
    mol2 = molecule.Molecule(eg.ch2f2[0], eg.ch2f2[2])
    bund = molecule.MoleculeBundle([mol1, mol2])
    stre = bund.measure('stre', 2, 3)
    assert np.allclose(stre, [np.sqrt(2), 2.04])


def test_MoleculeBundle_match_to_ref():
    mol1 = molecule.Molecule(eg.ch2f2[0], eg.ch2f2[1])
    mol2 = molecule.Molecule(eg.ch2f2[0], eg.ch2f2[2])
    mol3 = molecule.Molecule(eg.ch2f2[0], eg.ch2f2[3])
    mol4 = molecule.Molecule(eg.ch2f2[0], eg.ch2f2[4])
    mol5 = molecule.Molecule(eg.ch2f2[0], eg.ch2f2[5])
    bund1 = molecule.MoleculeBundle([mol3, mol4, mol5])
    bund2 = molecule.MoleculeBundle([mol1, mol2])
    inds = bund1.match_to_ref(bund2)
    assert np.allclose(inds, [0, 0, 1])


def test_import_molecule(tmpdir):
    f = tmpdir.join('tmp.xyz')
    f.write(ef.xyz_novec)
    mol = molecule.import_molecule(str(f.realpath()))
    assert np.all(mol.elem[1:] == eg.ch4[0])
    assert np.allclose(mol.xyz[1:], eg.ch4[1])


def test_import_bundle_single_file(tmpdir):
    f = tmpdir.join('tmp.xyz')
    f.write(ef.xyz_novec + ef.zmtvar_nocom)
    bund = molecule.import_bundle(str(f.realpath()))
    assert np.all(bund.molecules[0].elem[1:] == eg.ch4[0])
    assert np.all(bund.molecules[1].elem[1:] == eg.ch4[0])
    assert np.allclose(bund.molecules[0].xyz[1:], eg.ch4[1])
    assert np.allclose(bund.molecules[1].xyz[1:], eg.ch4_zmt[1])


def test_import_bundle_multiple_files(tmpdir):
    f1 = tmpdir.join('tmp1.xyz')
    f2 = tmpdir.join('tmp2.xyz')
    f1.write(ef.zmtvar_nocom + ef.zmtvar_nocom)
    f2.write(ef.xyz_novec)
    bund = molecule.import_bundle([str(f1.realpath()), str(f2.realpath())])
    assert np.all(bund.molecules[0].elem[1:] == eg.ch4[0])
    assert np.all(bund.molecules[1].elem[1:] == eg.ch4[0])
    assert np.all(bund.molecules[2].elem[1:] == eg.ch4[0])
    assert np.allclose(bund.molecules[0].xyz[1:], eg.ch4_zmt[1])
    assert np.allclose(bund.molecules[1].xyz[1:], eg.ch4_zmt[1])
    assert np.allclose(bund.molecules[2].xyz[1:], eg.ch4[1])



"""
Tests for the fileio module.
"""
import pytest
import numpy as np
import gimbal.fileio as fileio
from examples import Geometries as eg
from examples import Files as ef


def test_read_xyz_bohr(tmpdir):
    f = ef.tmpf(tmpdir, ef.xyz_bohr)
    elem, xyz, vec, com = fileio.read_xyz(f, units='bohr')
    assert np.all(elem == eg.ch4[0])
    assert np.allclose(xyz, eg.ch4[1])


def test_read_xyz_comment(tmpdir):
    f = ef.tmpf(tmpdir, ef.xyz_novec)
    elem, xyz, vec, com = fileio.read_xyz(f, hascom=True)
    assert np.all(elem == eg.ch4[0])
    assert np.allclose(xyz, eg.ch4[1])
    assert com == 'comment line'


def test_read_xyz_vector(tmpdir):
    f = ef.tmpf(tmpdir, ef.xyz_vec)
    elem, xyz, vec, com = fileio.read_xyz(f, hasvec=True)
    assert np.all(elem == eg.ch4[0])
    assert np.allclose(xyz, eg.ch4[1])
    assert np.allclose(vec, np.ones((5, 3)))


def test_read_xyz_vector_empty(tmpdir):
    f = ef.tmpf(tmpdir, ef.xyz_novec)
    elem, xyz, vec, com = fileio.read_xyz(f, hasvec=True)
    assert np.all(elem == eg.ch4[0])
    assert np.allclose(xyz, eg.ch4[1])
    assert np.allclose(vec, np.zeros((5, 3)))


def test_read_xyz_wrong_format(tmpdir):
    f = ef.tmpf(tmpdir, ef.col_nocom)
    with pytest.raises(IOError, match=r'geometry not in XYZ format.'):
        elem, xyz, vec, com = fileio.read_xyz(f)


def test_read_col_comment(tmpdir):
    f = ef.tmpf(tmpdir, ef.col_com)
    elem, xyz, vec, com = fileio.read_col(f, hascom=True)
    assert np.all(elem == eg.ch4[0])
    assert np.allclose(xyz, eg.ch4[1])
    assert com == 'comment line'


def test_read_col_hasvec(tmpdir):
    f = ef.tmpf(tmpdir, ef.col_nocom)
    elem, xyz, vec, com = fileio.read_col(f, hasvec=True)
    assert np.all(elem == eg.ch4[0])
    assert np.allclose(xyz, eg.ch4[1])
    assert np.allclose(vec, np.zeros_like(xyz))


def test_read_col_wrong_format(tmpdir):
    f = ef.tmpf(tmpdir, ef.gdat_bohr)
    with pytest.raises(IOError, match=r'geometry not in COLUMBUS format.'):
        elem, xyz, vec, com = fileio.read_col(f)


def test_read_gdat_comment(tmpdir):
    f = ef.tmpf(tmpdir, ef.gdat_bohr)
    elem, xyz, vec, com = fileio.read_gdat(f, hascom=True)
    assert np.all(elem == eg.ch4[0])
    assert np.allclose(xyz, eg.ch4[1])
    assert com == 'comment line'


def test_read_gdat_vector(tmpdir):
    f = ef.tmpf(tmpdir, ef.gdat_bohr)
    elem, xyz, vec, com = fileio.read_gdat(f, hasvec=True)
    assert np.all(elem == eg.ch4[0])
    assert np.allclose(xyz, eg.ch4[1])
    assert np.allclose(vec, np.ones((5, 3)))


def test_read_gdat_wrong_format(tmpdir):
    f = ef.tmpf(tmpdir, ef.zmt_nocom)
    with pytest.raises(IOError, match=r'geometry not in Geometry.dat .*'):
        elem, xyz, vec, com = fileio.read_gdat(f)


def test_read_zmt_comment(tmpdir):
    f = ef.tmpf(tmpdir, ef.zmt_com)
    elem, xyz, vec, com = fileio.read_zmt(f, hascom=True)
    assert np.all(elem == eg.ch4[0])
    assert np.allclose(xyz, eg.ch4_zmt[1], atol=1e-6)
    assert com == 'comment line'


def test_read_zmt_hasvec(tmpdir):
    f = ef.tmpf(tmpdir, ef.zmt_nocom)
    elem, xyz, vec, com = fileio.read_zmt(f, hasvec=True)
    assert np.all(elem == eg.ch4[0])
    assert np.allclose(xyz, eg.ch4_zmt[1], atol=1e-6)
    assert np.allclose(vec, np.zeros_like(xyz))


def test_read_zmt_var(tmpdir):
    f = ef.tmpf(tmpdir, ef.zmtvar_nocom)
    elem, xyz, vec, com = fileio.read_zmt(f)
    assert np.all(elem == eg.ch4_zmt[0])
    assert np.allclose(xyz, eg.ch4_zmt[1], atol=1e-6)


def test_read_zmt_wrong_format(tmpdir):
    f = ef.tmpf(tmpdir, ef.traj_nocom)
    with pytest.raises(IOError, match=r'geometry not in Z-matrix .*'):
        elem, xyz, vec, com = fileio.read_zmt(f)


def test_read_traj_no_elem(tmpdir):
    f = ef.tmpf(tmpdir, ef.traj_nocom)
    elem, xyz, vec, com = fileio.read_traj(f)
    assert np.all(elem == 5*['X'])
    assert np.allclose(xyz, eg.ch4[1], atol=1e-4)


def test_read_traj_comment(tmpdir):
    f = ef.tmpf(tmpdir, ef.traj_com)
    elem, xyz, vec, com = fileio.read_traj(f, hascom=True, elem=eg.ch4[0])
    assert np.all(elem == eg.ch4[0])
    assert np.allclose(xyz, eg.ch4[1], atol=1e-4)
    assert com == 'comment line'


def test_read_traj_vector(tmpdir):
    f = ef.tmpf(tmpdir, ef.traj_nocom)
    elem, xyz, vec, com = fileio.read_traj(f, hasvec=True, elem=eg.ch4[0])
    assert np.all(elem == eg.ch4[0])
    assert np.allclose(xyz, eg.ch4[1], atol=1e-4)
    assert np.allclose(vec, np.ones((5, 3)), atol=1e-4)


def test_read_traj_time(tmpdir):
    f = ef.tmpf(tmpdir, ef.traj_time)
    elem, xyz, vec, com = fileio.read_traj(f, hasvec=True, time=1.)
    assert np.all(elem == 5*['X'])
    assert np.allclose(xyz, eg.ch4[1], atol=1e-4)
    assert np.allclose(vec, np.ones((5, 3)), atol=1e-4)


def test_read_traj_autocom(tmpdir):
    f = ef.tmpf(tmpdir, ef.traj_nocom)
    elem, xyz, vec, com = fileio.read_traj(f, autocom=True)
    assert np.all(elem == 5*['X'])
    assert np.allclose(xyz, eg.ch4[1], atol=1e-4)
    assert com == 't=    0.00, state=   2, a^2=    0.2500'


def test_read_traj_empty_line(tmpdir):
    f = ef.tmpf(tmpdir, '\n')
    with pytest.raises(IOError, match=r'empty line provided'):
        elem, xyz, vec, com = fileio.read_traj(f)


def test_read_traj_wrong_format(tmpdir):
    f = ef.tmpf(tmpdir, ef.xyz_novec)
    with pytest.raises(IOError, match=r'geometry not in trajectory .*'):
        elem, xyz, vec, com = fileio.read_traj(f)


def test_read_auto_end_of_file(tmpdir):
    f = ef.tmpf(tmpdir, '')
    with pytest.raises(IOError, match=r'end of file'):
        elem, xyz, vec, com = fileio.read_auto(f)


def test_read_auto_only_comment(tmpdir):
    f = ef.tmpf(tmpdir, 'comment line\n')
    with pytest.raises(IOError, match=r'cannot have comment with single .*'):
        elem, xyz, vec, com = fileio.read_auto(f, hascom=True)


def test_read_auto_zmt_one_line(tmpdir):
    f = ef.tmpf(tmpdir, 'He\n')
    elem, xyz, vec, com = fileio.read_auto(f)
    assert np.all(elem == ['He'])
    assert np.allclose(xyz, np.zeros((1, 3)))


def test_read_auto_col_one_line(tmpdir):
    he_col = (' He    2.0    0.00000000    0.00000000    0.00000000' +
              '    4.00260325\n')
    f = ef.tmpf(tmpdir, he_col)
    elem, xyz, vec, com = fileio.read_auto(f)
    assert np.all(elem == ['He'])
    assert np.allclose(xyz, np.zeros((1, 3)))


def test_read_auto_traj_one_line(tmpdir):
    f = ef.tmpf(tmpdir, ef.traj_nocom)
    elem, xyz, vec, com = fileio.read_auto(f)
    assert np.all(elem == 5*['X'])
    assert np.allclose(xyz, eg.ch4[1], atol=1e-4)


def test_read_auto_unrecognized_one_line(tmpdir):
    f = ef.tmpf(tmpdir, '123 abc 456 def\n')
    with pytest.raises(IOError, match=r'single line input in unrecognized .*'):
        elem, xyz, vec, com = fileio.read_auto(f)


def test_read_auto_zmt_one_line_comment(tmpdir):
    f = ef.tmpf(tmpdir, 'comment line\nHe\n')
    elem, xyz, vec, com = fileio.read_auto(f, hascom=True)
    assert np.all(elem == ['He'])
    assert np.allclose(xyz, np.zeros((1, 3)))
    assert com == 'comment line'


def test_read_auto_col_one_line_comment(tmpdir):
    he_col = ('comment line\n' +
              ' He    2.0    0.00000000    0.00000000    0.00000000' +
              '    4.00260325\n')
    f = ef.tmpf(tmpdir, he_col)
    elem, xyz, vec, com = fileio.read_auto(f, hascom=True)
    assert np.all(elem == ['He'])
    assert np.allclose(xyz, np.zeros((1, 3)))
    assert com == 'comment line'


def test_read_auto_traj_one_line_comment(tmpdir):
    f = ef.tmpf(tmpdir, ef.traj_com)
    elem, xyz, vec, com = fileio.read_auto(f, hascom=True)
    assert np.all(elem == 5*['X'])
    assert np.allclose(xyz, eg.ch4[1], atol=1e-4)
    assert com == 'comment line'


def test_read_auto_unrecognized_one_line_comment(tmpdir):
    f = ef.tmpf(tmpdir, 'comment line\n123 abc 456 def\n')
    with pytest.raises(IOError, match=r'single line input in unrecognized .*'):
        elem, xyz, vec, com = fileio.read_auto(f, hascom=True)


def test_read_auto_xyz(tmpdir):
    f = ef.tmpf(tmpdir, ef.xyz_novec)
    elem, xyz, vec, com = fileio.read_auto(f)
    assert np.all(elem == eg.ch4[0])
    assert np.allclose(xyz, eg.ch4[1])


def test_read_auto_col(tmpdir):
    f = ef.tmpf(tmpdir, ef.col_nocom)
    elem, xyz, vec, com = fileio.read_auto(f)
    assert np.all(elem == eg.ch4[0])
    assert np.allclose(xyz, eg.ch4[1])


def test_read_auto_col_comment(tmpdir):
    f = ef.tmpf(tmpdir, ef.col_com)
    elem, xyz, vec, com = fileio.read_auto(f, hascom=True)
    assert np.all(elem == eg.ch4[0])
    assert np.allclose(xyz, eg.ch4[1])
    assert com == 'comment line'


def test_read_auto_gdat(tmpdir):
    f = ef.tmpf(tmpdir, ef.gdat_bohr)
    elem, xyz, vec, com = fileio.read_auto(f)
    assert np.all(elem == eg.ch4[0])
    assert np.allclose(xyz, eg.ch4[1])


def test_read_auto_zmt(tmpdir):
    f = ef.tmpf(tmpdir, ef.zmt_nocom)
    elem, xyz, vec, com = fileio.read_auto(f)
    assert np.all(elem == eg.ch4[0])
    assert np.allclose(xyz, eg.ch4_zmt[1], atol=1e-6)


def test_read_auto_zmt_comment(tmpdir):
    f = ef.tmpf(tmpdir, ef.zmt_com)
    elem, xyz, vec, com = fileio.read_auto(f, hascom=True)
    assert np.all(elem == eg.ch4[0])
    assert np.allclose(xyz, eg.ch4_zmt[1], atol=1e-6)
    assert com == 'comment line'


def test_read_auto_traj(tmpdir):
    f = ef.tmpf(tmpdir, ef.traj_time)
    elem, xyz, vec, com = fileio.read_auto(f)
    assert np.all(elem == 5*['X'])
    assert np.allclose(xyz, eg.ch4[1], atol=1e-4)


def test_read_auto_unrecognized_multi_lines(tmpdir):
    f = ef.tmpf(tmpdir, 'abc 123 def 456\nabc def 123 456\n123 456 abc def\n')
    with pytest.raises(IOError, match=r'unrecognized file format'):
        elem, xyz, vec, com = fileio.read_auto(f)


def test_read_multiple_no_geoms(tmpdir):
    f = ef.tmpf(tmpdir, '\n')
    with pytest.raises(IOError, match=r'no geometries read from input files'):
        moldat = fileio.read_multiple(f)


def test_write_xyz_vec(tmpdir):
    f = tmpdir.join('tmp.geom')
    fileio.write_xyz(f.open(mode='w'), eg.ch4[0], eg.ch4[1],
                     vec=np.ones((5, 3)))
    assert f.read() == ef.xyz_vec.replace('comment line', '')


def test_write_xyz_bohr(tmpdir):
    f = tmpdir.join('tmp.geom')
    fileio.write_xyz(f.open(mode='w'), eg.ch4[0], eg.ch4[1], units='bohr')
    assert f.read() == ef.xyz_bohr.replace('comment line', '')


def test_write_col_comment(tmpdir):
    f = tmpdir.join('tmp.geom')
    fileio.write_col(f.open(mode='w'), eg.ch4[0], eg.ch4[1],
                     comment='comment line')
    assert f.read() == ef.col_com


def test_write_gdat_vec(tmpdir):
    f = tmpdir.join('tmp.geom')
    fileio.write_gdat(f.open(mode='w'), eg.ch4[0], eg.ch4[1],
                      vec=np.ones((5, 3)))
    assert f.read() == ef.gdat_bohr.replace('comment line', '')


def test_write_zmt_comment(tmpdir):
    f = tmpdir.join('tmp.geom')
    fileio.write_zmt(f.open(mode='w'), eg.ch4[0], eg.ch4[1],
                     comment='comment line')
    assert f.read() == ef.zmt_com


def test_write_zmtvar_comment(tmpdir):
    f = tmpdir.join('tmp.geom')
    fileio.write_zmtvar(f.open(mode='w'), eg.ch4[0], eg.ch4[1],
                        comment='comment line')
    assert f.read() == ef.zmtvar_com


def test_write_traj_vec(tmpdir):
    f = tmpdir.join('tmp.geom')
    fileio.write_traj(f.open(mode='w'), eg.ch4[0], eg.ch4[1],
                      vec=np.ones((5, 3)))
    soln = ef.traj_nocom.replace('0.4000    0.3000    0.2500    2.0000',
                                 '0.0000    0.0000    0.0000    0.0000')
    assert f.read() == soln


def test_write_traj_comment(tmpdir):
    f = tmpdir.join('tmp.geom')
    fileio.write_traj(f.open(mode='w'), eg.ch4[0], eg.ch4[1],
                      comment='comment line')
    soln = ef.traj_com.replace('0.4000    0.3000    0.2500    2.0000',
                               '0.0000    0.0000    0.0000    0.0000')
    assert f.read() == soln.replace(' 1.0000', ' 0.0000')


def test_write_auto_col(tmpdir):
    f = tmpdir.join('tmp.col')
    fileio.write_auto(f.open(mode='w'), eg.ch4[0], eg.ch4[1])
    assert f.read() == ef.col_nocom


def test_write_auto_gdat(tmpdir):
    f = tmpdir.join('tmp.dat')
    fileio.write_auto(f.open(mode='w'), eg.ch4[0], eg.ch4[1])
    soln = ef.gdat_bohr.replace('comment line', '').replace(' 1.000', ' 0.000')
    assert f.read() == soln


def test_write_auto_zmt(tmpdir):
    f = tmpdir.join('tmp.zmt')
    fileio.write_auto(f.open(mode='w'), eg.ch4[0], eg.ch4[1])
    assert f.read() == ef.zmt_nocom


def test_write_auto_zmtvar(tmpdir):
    f = tmpdir.join('tmp.zmtvar')
    fileio.write_auto(f.open(mode='w'), eg.ch4[0], eg.ch4[1])
    assert f.read() == ef.zmtvar_nocom


def test_write_auto_traj(tmpdir):
    f = tmpdir.join('tmp.tj')
    fileio.write_traj(f.open(mode='w'), eg.ch4[0], eg.ch4[1])
    soln = ef.traj_nocom.replace('0.4000    0.3000    0.2500    2.0000',
                                 '0.0000    0.0000    0.0000    0.0000')
    assert f.read() == soln.replace(' 1.0000', ' 0.0000')


def test_write_auto_xyz(tmpdir):
    f = tmpdir.join('tmp.xyz')
    fileio.write_auto(f.open(mode='w'), eg.ch4[0], eg.ch4[1])
    assert f.read() == ef.xyz_novec.replace('comment line', '')


def test_convert_xyz_to_col(tmpdir):
    f1 = tmpdir.join('tmp.xyz')
    f1.write(ef.xyz_novec)
    f2 = tmpdir.join('tmp.col')
    fileio.convert(str(f1.realpath()), str(f2.realpath()),
                   infmt='xyz', outfmt='col')
    assert f2.read() == ef.col_nocom


def test_convert_xyz_to_zmt_inunits_bohr(tmpdir):
    f1 = tmpdir.join('tmp.xyz')
    f1.write(ef.xyz_bohr.replace('88861', '8886072'))
    f2 = tmpdir.join('tmp.col')
    fileio.convert(str(f1.realpath()), str(f2.realpath()), inunits='bohr')
    assert f2.read() == ef.col_nocom


def test_convert_xyz_to_gdat_outunits_angstrom(tmpdir):
    f1 = tmpdir.join('tmp.xyz')
    f1.write(ef.xyz_vec)
    f2 = tmpdir.join('tmp.dat')
    fileio.convert(str(f1.realpath()), str(f2.realpath()), outunits='ang',
                   hasvec=True)
    assert f2.read() == ef.gdat_ang


def test_convert_xyz_to_zmtvar_hascom(tmpdir):
    f1 = tmpdir.join('tmp.xyz')
    f1.write(ef.xyz_novec)
    f2 = tmpdir.join('tmp.zmtvar')
    fileio.convert(str(f1.realpath()), str(f2.realpath()), hascom=True)
    assert f2.read() == ef.zmtvar_com


def test_convert_xyz_to_traj_hasvec(tmpdir):
    f1 = tmpdir.join('tmp.xyz')
    f1.write(ef.xyz_vec)
    f2 = tmpdir.join('tmp.tj')
    fileio.convert(str(f1.realpath()), str(f2.realpath()), hasvec=True)
    soln = ef.traj_nocom.replace('0.4000    0.3000    0.2500    2.0000',
                                 '0.0000    0.0000    0.0000    0.0000')
    assert f2.read() == soln


def test_get_optargs_no_default():
    args = ['-u', 'arg1', 'arg2', '-v']
    val1 = fileio.get_optarg(args, '-u')
    val2 = fileio.get_optarg(args, '-v')
    assert val1 is True
    assert val2 is True
    assert args == ['arg1', 'arg2']


def test_get_optargs_set_default():
    args = ['-u', 'optarg1', 'arg1', 'arg2', '-v', 'optarg2']
    val1 = fileio.get_optarg(args, '-u', default=None)
    val2 = fileio.get_optarg(args, '-v', default=None)
    assert val1 == 'optarg1'
    assert val2 == 'optarg2'
    assert args == ['arg1', 'arg2']


def test_get_optargs_not_found():
    args = ['arg1', 'arg2']
    val1 = fileio.get_optarg(args, '-u')
    val2 = fileio.get_optarg(args, '-v', default=None)
    assert val1 is False
    assert val2 is None
    assert args == ['arg1', 'arg2']


def test_valvar_fails(tmpdir):
    tdict = dict(a = 1.)
    with pytest.raises(KeyError, match=r'.* not found in variable list'):
        fileio._valvar('b', tdict)



"""
Tests for the substitute module.
"""
import pytest
import numpy as np
import gimbal.substitute as substitute
from examples import Geometries as eg


def test_add_subs_fails():
    lib = substitute.SubLib()
    with pytest.raises(ValueError, match=r'Number of inds != number of .*'):
        elem, xyz = lib.add_subs(['me', 'me', 'me'], inds=[1, 2])


def test_import_sub_methyl():
    lib = substitute.SubLib()
    elem, xyz = substitute.import_sub('ch3')
    assert np.all(elem == lib.elem['me'])
    assert np.allclose(xyz, lib.xyz['me'])


def test_import_sub_ethyl():
    lib = substitute.SubLib()
    elem, xyz = substitute.import_sub('ch2ch3')
    assert np.all(elem == lib.elem['et'])
    assert np.allclose(xyz, lib.xyz['et'])


def test_subst_fluoro_default():
    elem, xyz, vec = substitute.subst(eg.c2h4[0], eg.c2h4[1], 'f', 2)
    assert np.all(elem == eg.c2h3f[0])
    assert np.allclose(xyz, eg.c2h3f[1])
    assert vec is None


def test_subst_fluoro_multisub():
    elem, xyz, vec = substitute.subst(eg.c2h4[0], eg.c2h4[1], 'f', [2, 3])
    assert np.all(elem == np.hstack((eg.c2h3f[0][:3], eg.c2h3f[0][4:])))
    assert np.allclose(xyz, np.vstack((eg.c2h3f[1][:3], eg.c2h3f[1][4:])))


def test_subst_fluoro_with_fluoro():
    inp_xyz = np.copy(eg.c2h3f[1])
    inp_xyz[2] += np.ones(3)
    elem, xyz, vec = substitute.subst(eg.c2h3f[0], inp_xyz, 'f', 2)
    assert np.all(elem == eg.c2h3f[0])
    assert np.allclose(xyz, inp_xyz)


def test_subst_fluoro_vec():
    vec = np.ones_like(eg.c2h4[1])
    elem, xyz, vec = substitute.subst(eg.c2h4[0], eg.c2h4[1], 'f', 2, vec=vec)
    soln = np.copy(vec)
    soln[2] = 0.
    assert np.all(elem == eg.c2h3f[0])
    assert np.allclose(xyz, eg.c2h3f[1])
    assert np.allclose(vec, soln)


def test_subst_fluoro_ibond():
    elem, xyz, vec = substitute.subst(eg.c2h4[0], eg.c2h4[1], 'f', 2, ibond=0)
    assert np.all(elem == eg.c2h3f[0])
    assert np.allclose(xyz, eg.c2h3f[1])


def test_subst_methyl_plane_ind():
    elem, xyz, vec = substitute.subst(eg.c2h4[0], eg.c2h4[1], 'ch3', 2, pl=1)
    assert np.all(elem == eg.c2h3me[0])
    assert np.allclose(xyz, eg.c2h3me[1])


def test_subst_methyl_plane_vec():
    elem, xyz, vec = substitute.subst(eg.c2h4[0], eg.c2h4[1], 'ch3', 2,
                                      pl=[1, 1, 1])
    assert np.all(elem == eg.c2h3me_ax[0])
    assert np.allclose(xyz, eg.c2h3me_ax[1])


def test_subst_ibond_equals_ipos():
    with pytest.raises(ValueError, match=r'sub and bond indices .*'):
        elem, xyz, vec = substitute.subst(eg.c2h4[0], eg.c2h4[1], 'f', 2,
                                          ibond=2)


def test_subst_pl_equals_ipos():
    with pytest.raises(ValueError, match=r'plane and sub indices .*'):
        elem, xyz, vec = substitute.subst(eg.c2h4[0], eg.c2h4[1], 'f', 2, pl=2)


def test_subst_pl_equals_ibond():
    with pytest.raises(ValueError, match=r'plane and bond indices .*'):
        elem, xyz, vec = substitute.subst(eg.c2h4[0], eg.c2h4[1], 'f', 2,
                                          ibond=3, pl=3)



"""
Routines to detect, symmetrize and generate symmetric geometries.

The symmetric axes are automatically found by the principal axes of
rotation for convenience. As a result, some axes need to be rearranged
to match normal conventions depending on the type of rotor.
"""
import numpy as np
import gimbal.constants as con
import gimbal.displace as displace


def principal_axes(elem, xyz):
    """Returns the moments of inertia and corresponding principal axes of
    a geometry.

    The moments of inertia are found by diagonalizing the inertia tensor,

    .. math::

        \mathbf{A} = (\mathbf{r} \cdot \mathbf{r}) \mathbf{I} -
        \mathbf{r} \otimes \mathbf{r}

    where **I** is the identity and :math:`\otimes` is an open product.
    The coordinates **r** are mass weighted cartesian coordinates,
    :math:`r_i = m_i^{1/2} q_i`

    Parameters
    ----------
    elem : (N,) array_like
        The atomic element list.
    xyz : (N, 3) array_like
        The cartesian coordinates of each atom.

    Returns
    -------
    (3,) ndarray
        The magnitudes of principal moments of inertia.
    (3, 3) ndarray
        The principal axes of inertia.
    """
    rxyz = np.sqrt(con.get_mass(elem))[:,np.newaxis] * xyz
    inert = np.sum(rxyz**2) * np.eye(3) - rxyz.T.dot(rxyz)
    return np.linalg.eig(inert)


def assign_rotor(moms):
    """Returns the type of rotor based on the moments of inertia."""
    pass


def symmetrize(elem, xyz, thresh=1e-3):
    """Returns a geometry and its point group with symmetry defined by a
    geometric threshold.

    The output geometry is rotated to have symmetry elements along
    cartesian axes if possible.

    Parameters
    ----------
    elem : (N,) array_like
        The atomic element list.
    xyz : (N, 3) array_like
        The cartesian coordinates of each atom.
    thresh : float, optional
        The tolerance threshold for symmetry recognition.

    Returns
    -------
    (N, 3) ndarray
        The cartesian coordinates of the symmetrized molecule.
    """
    # centre and align
    new_xyz = displace.centre_mass(elem, xyz)
    moms, eigv = principal_axes(elem, new_xyz)
    eigv[:,-1] *= np.linalg.det(eigv)
    new_xyz = new_xyz.dot(eigv)

    # find atoms of the same type that are equidistant from the origin
    dist = np.linalg.norm(new_xyz, axis=1)
    #...
    return new_xyz


def cart_jumble(xyz, thresh=1e-3):
    """Moves cartesian coordinates by random amounts up to a given
    threshold.

    Parameters
    ----------
    xyz : (N, 3) array_like
        The set of atomic cartesian coordinates.
    thresh : float, optional
        The maximum random cartesian displacement.

    Returns
    -------
    (N, 3) ndarray
        The randomly displaced cartesian coordinates.
    """
    return xyz + thresh*np.random.rand(*xyz.shape)



r"""
Routines for displacing a molecular geometry by translation or
proper/improper rotation.
::

        1
        |
        4
       / \
      2   3

Example axes for displacements:

    1. X1X4 stretch: :math:`r_{14}`
    2. X1X4 torsion: :math:`r_{14}` (for motion of 2, 3)
    3. X1X4X2 bend: :math:`r_{14} \times r_{24}`
    4. X1 out-of-plane: :math:`r_{24} - r_{34}` or
       :math:`(r_{24} \times r_{34}) \times r_{14}`

Each internal coordinate measurement has the option of changing the units
(see the constants module) or taking the absolute value.
"""
import operator as op
import pyparsing as pp
import numpy as np
import gimbal.constants as con


class VectorParser(object):
    """An object for defining and evaluating vector operations on
    a cartesian geometry.

    A new VectorParser instance takes a cartesian geometry as an
    optional input. The instance can be called with a vector (no
    action), 3x3 array (cross product) or string, parsed according
    to the syntax in :func:`~VectorParser.generate_parser`.

    Attributes
    ----------
    xyz : (N, 3) array_like, optional
        The cartesian geometry which defines the indices in parsed
        expressions. If None, only expressions without indices can
        be parsed.
    unop : dict
        A dictionary which defines unary operations.
    bnadd : dict
        A dictionary which defines binary addition operations.
    bnmul : dict
        A dictionary which defines binary multiplication operations.
    bnop : dict
        A dictionary which defines all binary operations.
    axes : dict
        A dictionary which defines cartesian axis labels.
    expr : pyparsing.Forward
        A pyparsing grammar used to evaluate expressions. Automatically
        generated when xyz is set.
    """
    def __init__(self, xyz=None):
        self.unop = {'+': op.pos, '-': op.neg}
        self.bnadd = {'+': op.add, '-': op.sub}
        self.bnmul = {'*': op.mul, '/': op.truediv, 'o': np.dot, 'x': np.cross}
        self.bnop = self.bnadd.copy()
        self.bnop.update(self.bnmul)
        self.axes = dict(X = np.array([1., 0., 0.]),
                         Y = np.array([0., 1., 0.]),
                         Z = np.array([0., 0., 1.]))
        self.xyz = xyz

    def __call__(self, inp, unit=False):
        """Evaluates an expression based on a string.

        Parameters
        ----------
        inp : str or array_like
            A string or array used to specify an axis.
        unit : bool, optional
            Specifies if the axis is converted to a unit vector.

        Returns
        -------
        float or ndarray
            The result of the vector operation.

        Raises
        ------
        ValueError
            If input is not a string, 3-vector or 3x3 array.
        """
        if isinstance(inp, str):
            u = self.expr.parseString(inp, parseAll=True)[0]
        elif len(inp) == 3:
            u = np.array(inp, dtype=float)
            if u.size == 9:
                u = np.cross(u[0] - u[1], u[2] - u[1])
        else:
            raise ValueError('Axis specification not recognized')

        if unit:
            return con.unit_vec(u)
        else:
            return u

    @property
    def xyz(self):
        """Gets the value of xyz."""
        return self._xyz

    @xyz.setter
    def xyz(self, val):
        """Sets the value of xyz and generates the parser."""
        self.expr = self.generate_parser(val)
        self._xyz = val

    def _eval_unary(self, tokens):
        """Evaluates unary operations.

        Parameters
        ----------
        tokens : list
            A list of pyparsing tokens from a matching unary expression.

        Returns
        -------
        float or ndarray
            The expression after unary operation.
        """
        vals = tokens[0]
        return self.unop[vals[0]](vals[1])

    def _eval_binary(self, tokens):
        """Evaluates binary operations.

        Parameters
        ----------
        tokens : list
            A list of pyparsing tokens from a matching binary expression.

        Returns
        -------
        float or ndarray
            The expression after binary operation.
        """
        vals = tokens[0]
        newval = vals[0]
        it = iter(vals[1:])
        for oper in it:
            newval = self.bnop[oper](newval, next(it))

        return newval

    def _eval_power(self, tokens):
        """Evaluates power operations.

        Parameters
        ----------
        tokens : list
            A list of pyparsing tokens from a matching power expression.

        Returns
        -------
        float or ndarray
            The expression after power operation.
        """
        vals = tokens[0]
        newval = vals[-1]
        for v in vals[-3::-2]:
            newval = v**newval

        return newval

    def generate_parser(self, xyz=None):
        """Creates the pyparsing expression based on geometry.

        The syntax is as follows:

            - ``i+`` are indices of xyz and return vectors.
            - ``i+.j`` are floating point numbers (j optional).
            - ``i[j]`` is the j-th (scalar) element of xyz[i].
            - ``X, Y, Z`` are unit vectors along x, y and z axes (uppercase only).
            - ``+`` and ``-`` are addition/subtraction of vectors or scalars.
            - ``*`` and ``/`` are multiplication/division of vectors and scalars
              (elementwise).
            - ``o`` and ``x`` are scalar/vector products of vectors only.
            - ``^`` is the power of a vector/scalar by a scalar (elementwise).
            - ``(`` and ``)`` specify order of operation.
            - ``[i, j, k]`` gives a vector with scalar elements i, j and k.

        Parameters
        ----------
        xyz : (N, 3), array_like, optional
            The cartesian geometry used in index expressions. If not
            provided, strings containing indices will raise an error.

        Returns
        -------
        pyparsing.Forward
            A pyparsing grammar definition.
        """
        expr = pp.Forward()

        # operand types: int, int with index, float, axis or delimited list
        intnum = pp.Word(pp.nums)
        fltind = pp.Word(pp.nums) + '[' + pp.Word(pp.nums) + ']'
        fltnum = pp.Combine(pp.Word(pp.nums) + '.' + pp.Optional(pp.Word(pp.nums)))
        alphax = pp.oneOf(' '.join(self.axes))
        dllist = pp.Suppress('[') + pp.delimitedList(expr) + pp.Suppress(']')
        intnum.setParseAction(lambda t: xyz[int(t[0])])
        fltind.setParseAction(lambda t: xyz[int(t[0])][int(t[2])])
        fltnum.setParseAction(lambda t: float(t[0]))
        alphax.setParseAction(lambda t: self.axes[t[0]])
        dllist.setParseAction(lambda t: np.array(t[:]))
        operand = dllist | alphax | fltnum | fltind | intnum

        # operators: unary, power, binary multiplication/division,
        #    binary addition/subtraction
        sgnop = pp.oneOf(' '.join(self.unop))
        expop = pp.Literal('^')
        mulop = pp.oneOf(' '.join(self.bnmul))
        addop = pp.oneOf(' '.join(self.bnadd))

        # set operator precedence
        prec = [(sgnop, 1, pp.opAssoc.RIGHT, self._eval_unary),
                (expop, 2, pp.opAssoc.LEFT, self._eval_power),
                (mulop, 2, pp.opAssoc.LEFT, self._eval_binary),
                (addop, 2, pp.opAssoc.LEFT, self._eval_binary)]
        return expr << pp.infixNotation(operand, prec)


def translate(xyz, amp, axis, ind=None, units='ang'):
    """Translates a set of atoms along a given vector.

    Parameters
    ----------
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.
    amp : float
        The distance for translation.
    axis : array_like or str
        The axis of translation, parsed by :class:`VectorParser`.
    ind : array_like, optional
        List of atomic indices to specify which atoms are displaced. If
        ind is None (default) then all atoms are displaced.
    units : str, optional
        The units of length for displacement. Default is angstroms.

    Returns
    -------
    (N, 3) ndarray
        The atomic cartesian coordinates of the displaced molecule.
    """
    if ind is None:
        ind = range(len(xyz))
    vp = VectorParser(xyz)
    u = vp(axis, unit=True)
    amp *= con.conv(units, 'ang')

    newxyz = np.copy(xyz)
    newxyz[ind] += amp * u
    return newxyz


def rotmat(ang, u, det=1, units='rad', xyz=None):
    r"""Returns the rotational matrix based on an angle and axis.

    A general rotational matrix in 3D can be formed given an angle and
    an axis by

    .. math::

        \mathbf{R} = \cos(a) \mathbf{I} + (\det(\mathbf{R}) -
        \cos(a)) \mathbf{u} \otimes \mathbf{u} + \sin(a) [\mathbf{u}]_\times

    for identity matrix **I**, angle *a*, axis **u**,
    outer product :math:`\otimes` and cross-product matrix
    :math:`[\mathbf{u}]_\times`.  Determinants of +1 and -1 give proper and
    improper rotation, respectively. Thus, :math:`\det(\mathbf{R}) = -1` and
    :math:`a = 0` is a reflection along the axis. Action of the rotational
    matrix occurs about the origin. See en.wikipedia.org/wiki/Rotation_matrix
    and http://scipp.ucsc.edu/~haber/ph251/rotreflect_17.pdf

    Parameters
    ----------
    ang : float
        The angle of rotation.
    u : array_like or str
        The axis of rotation, converted to a unit vector.
    det : int, optional
        The determinant of the matrix (1 or -1) used to specify proper
        and improper rotations. Default is 1.
    units : str, optional
        The units of angle for the rotation. Default is radians.
    xyz : (N, 3) array_like, optional
        The cartesian coordinates used in axis specification.

    Returns
    -------
    (3, 3) ndarray
        The rotational matrix of the given angle and axis.

    Raises
    ------
    ValueError
        When the absolute value of the determinant is not equal to 1.
    """
    if not np.isclose(np.abs(det), 1):
        raise ValueError('Determinant of a rotational matrix must be +/- 1')

    u /= np.linalg.norm(u)
    amp = ang * con.conv(units, 'rad')
    ucross = np.array([[0, u[2], -u[1]], [-u[2], 0, u[0]], [u[1], -u[0], 0]])
    return (np.cos(amp) * np.eye(3) + np.sin(amp) * ucross +
            (det - np.cos(amp)) * np.outer(u, u))


def angax(rotmat, units='rad'):
    r"""Returns the angle, axis of rotation and determinant of a
    rotational matrix.

    Based on the form of **R**, it can be separated into symmetric
    and antisymmetric components with :math:`(r_{ij} + r_{ji})/2` and
    :math:`(r_{ij} - r_{ji})/2`, respectively. Then,

    .. math::

        r_{ii} = \cos(a) + u_i^2 (\det(\mathbf{R}) - \cos(a)),

        \cos(a) = (-\det(\mathbf{R}) + \sum_j r_{jj}) / 2 = 
        (\mathrm{tr}(\mathbf{R}) - \det(\mathbf{R})) / 2.

    From the expression for :math:`r_{ii}`, the magnitude of :math:`u_i`
    can be found

    .. math::

        |u_i| = \sqrt{\frac{1 + \det(\mathbf{R}) [2 r_{ii} -
        \mathrm{tr}(\mathbf{R})])}{3 - \det(\mathbf{R}) \mathrm{tr}(\mathbf{R})}},

    which satisfies :math:`u \cdot u = 1`. Note that if
    :math:`\det(\mathbf{R}) \mathrm{tr}(\mathbf{R}) = 3`, the axis is arbitrary
    (identity or inversion). Otherwise, the sign can be found from the
    antisymmetric component of **R**.

    .. math::

        u_i \sin(a) = (r_{jk} - r_{kj}) / 2, \quad i \neq j \neq k,

        \mathrm{sign}(u_i) = \mathrm{sign}(r_{jk} - r_{kj}),

    since :math:`\sin(a)` is positive in the range 0 to :math:`\pi`. :math:`i`,
    :math:`j` and :math:`k` obey the cyclic relation 3 -> 2 -> 1 -> 3 -> ...

    This fails when :math:`det(\mathbf{R}) \mathrm{tr}(\mathbf{R}) = -1`, in
    which case the symmetric component of **R** is used

    .. math::

        u_i u_j (\det(\mathbf{R}) - \cos(a)) = (r_{ij} + r_{ji}) / 2,

        \mathrm{sign}(u_i) \mathrm{sign}(u_j) = \det(\mathbf{R}) \mathrm{sign}(r_{ij} + r_{ji}).

    The signs can then be found by letting :math:`\mathrm{sign}(u_3) = +1`,
    since a rotation of :math:`pi` or a reflection are equivalent for
    antiparallel axes. See
    http://scipp.ucsc.edu/~haber/ph251/rotreflect_17.pdf

    Parameters
    ----------
    rotmat : (3, 3) array_like
        The rotational matrix.
    units : str, optional
        The output units for the angle. Default is radians.

    Returns
    -------
    ang : float
        The angle of rotation.
    u : (3,) ndarray
        The axis of rotation as a 3D vector.
    det : int
        The determinant of the rotation matrix.

    Raises
    ------
    ValueError
        When the absolute value of the determinant is not equal to 1.
    """
    det = np.linalg.det(rotmat)
    if not np.isclose(np.abs(det), 1):
        raise ValueError('Determinant of a rotational matrix must be +/- 1')

    tr = np.trace(rotmat)
    ang = con.arccos((tr - det) / 2) * con.conv('rad', units)
    if np.isclose(det*tr, 3):
        u = np.array([0, 0, 1])
    else:
        u = np.sqrt((1 + det*(2*np.diag(rotmat) - tr)) / (3 - det*tr))
        if np.isclose(det*tr, -1):
            sgn = np.ones(3)
            sgn[1] = det * _nonzero_sign(rotmat[1,2] + rotmat[2,1])
            sgn[0] = det * sgn[1] * _nonzero_sign(rotmat[0,1] + rotmat[1,0])
            u *= sgn
        else:
            u[0] *= _nonzero_sign(rotmat[1,2] - rotmat[2,1])
            u[1] *= _nonzero_sign(rotmat[2,0] - rotmat[0,2])
            u[2] *= _nonzero_sign(rotmat[0,1] - rotmat[1,0])

    return ang, u, det


def rotate(xyz, ang, axis, ind=None, origin=np.zeros(3), det=1, units='rad'):
    """Rotates a set of atoms about a given vector.

    An origin can be specified for rotation about a specific point. If
    no indices are specified, all atoms are displaced. Setting det=-1
    leads to an improper rotation.

    Parameters
    ----------
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.
    ang : float
        The angle of rotation.
    axis : array_like or str
        The axis of rotation, parsed by :class:`VectorParser`.
    ind : array_like, optional
        List of atomic indices to specify which atoms are displaced. If
        ind is None (default) then all atoms are displaced.
    origin : (3,) array_like, optional
        The origin of rotation, parsed by :class:`VectorParser`. Default
        is the cartesian origin.
    det : float, optional
        The determinant of the rotation. 1 (default) is a proper rotation
        and -1 is an improper rotation (rotation + reflection).
    units : str, optional
        The units of length for displacement. Default is angstroms.

    Returns
    -------
    (N, 3) ndarray
        The atomic cartesian coordinates of the displaced molecule.
    """
    if ind is None:
        ind = range(len(xyz))

    vp = VectorParser(xyz)
    ax = vp(axis, unit=True)
    org = vp(origin)
    newxyz = xyz - org
    newxyz[ind] = newxyz[ind].dot(rotmat(ang, ax, det=det, units=units))
    return newxyz + org


def align_pos(xyz, test_crd, ref_crd, ind=None):
    """Translates a set of atoms such that two positions are coincident.

    Parameters
    ----------
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.
    test_crd : (3,) array_like
        Cartesian coordinates of the original position.
    test_crd : (3,) array_like
        Cartesian coordinates of the final position.
    ind : array_like, optional
        List of atomic indices to specify which atoms are displaced. If
        `ind == None` (default) then all atoms are displaced.

    Returns
    -------
    (N, 3) ndarray
        The atomic cartesian coordinates of the displaced molecule.
    """
    transax = ref_crd - test_crd
    dist = np.linalg.norm(transax)
    return translate(xyz, dist, transax, ind=ind)


def align_axis(xyz, test_ax, ref_ax, ind=None, origin=np.zeros(3)):
    """Rotates a set of atoms such that two axes are parallel.

    Parameters
    ----------
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.
    test_crd : (3,) array_like
        Cartesian coordinates of the original axis.
    test_crd : (3,) array_like
        Cartesian coordinates of the final axis.
    ind : array_like, optional
        List of atomic indices to specify which atoms are displaced. If
        `ind == None` (default) then all atoms are displaced.
    origin : (3,) array_like, optional
        The origin of rotation. Default is the cartesian origin.

    Returns
    -------
    (N, 3) ndarray
        The atomic cartesian coordinates of the displaced molecule.
    """
    vp = VectorParser(xyz)
    test = vp(test_ax, unit=True)
    ref = vp(ref_ax, unit=True)
    if np.allclose(test, ref):
        return xyz
    elif np.allclose(test, -ref):
        rotax = np.array([0., 0., 1.])
        if np.allclose(test, rotax) or np.allclose(test, -rotax):
            rotax = np.array([0., 1., 0.])
        rotax -= np.dot(rotax, test) * test
        return rotate(xyz, np.pi, rotax, ind=ind, origin=origin)
    else:
        angle = con.arccos(np.dot(test, ref))
        rotax = np.cross(test, ref)
        return rotate(xyz, angle, rotax, ind=ind, origin=origin)


def get_centremass(elem, xyz):
    """Returns centre of mass of a set of atoms.

    Parameters
    ----------
    elem : (N,) array_like
        The atomic symbols.
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.

    Returns
    -------
    (3,) ndarray
        The position of the centre of mass.
    """
    mass = con.get_mass(elem)
    if isinstance(mass, float):
        # Centre of mass of one atom is its position
        return xyz
    elif np.allclose(mass, mass[0]):
        # If masses are identical (including zero), return mean position
        return np.sum(xyz, axis=0) / len(xyz)
    else:
        return np.sum(mass[:,np.newaxis] * xyz, axis=0) / np.sum(mass)


def centre_mass(elem, xyz):
    """Returns xyz with centre of mass at the origin.

    If an index list is provided to inds, only the centre of mass of
    atoms at those indices will be used.

    Parameters
    ----------
    elem : (N,) array_like
        The atomic symbols.
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.

    Returns
    -------
    (N, 3) ndarray
        The atomic cartesian coordinates of the displaced molecule.
    """
    return xyz - get_centremass(elem, xyz)


def _nonzero_sign(x):
    """Returns the sign of a nonzero number, otherwise returns 1.

    Parameters
    ----------
    x : float or int
        The input number.

    Returns
    -------
    float
        The sign of x, i.e. +1 or -1.
    """
    if np.isclose(x, 0.):
        return 1.
    else:
        return np.sign(x)



"""
Substitution of molecular geometries with functional groups.

Given a cartesian geometry and element labels, a substituent can
be added knowing (a) the substituent identity (b) the desired
position to substitute and (c) the bond axis for substitution.

This requires some default information, such as the default structure
and orientation of the substituent (relative to an axis) and the
bond length of the substituent. For now, only single bonds are treated.
"""
import numpy as np
import gimbal.displace as displace
import gimbal.fileio as fileio
import gimbal.constants as con


class SubLib(object):
    """
    Object containing a library of substituent geometries.

    Attributes
    ----------
    syn : dict
        A dictionary of synonyms used to find the appropriate
        substituents.
    elem : dict
        A dictionary of atomic symbols for each substituent.
    xyz : dict
        A dictionary of atomic cartesian coordinates for each
        substituent.
    """
    def __init__(self):
        self.syn = dict()
        self.elem = dict()
        self.xyz = dict()
        self._populate_syn()
        self._populate_elem()
        self._populate_xyz()
        self._add_comb()

    def _populate_syn(self):
        """Adds a dictionary of synonyms for labels."""
        synlist = [['h'],
                   ['d'],
                   ['me', 'ch3', 'h3c'],
                   ['et', 'ch2ch3', 'c2h5', 'ch3ch2'],
                   ['npr', 'ch2ch2ch3', 'c3h7', 'ch3ch2ch2'],
                   ['ipr', 'chch3ch3', 'ch(ch3)2', '(ch3)2ch', 'ch3chch3'],
                   ['nbu', 'ch2ch2ch2ch3', 'c4h9', 'ch3ch2ch2ch2'],
                   ['ibu', 'chch3ch2ch3', 'ch3chch2ch3', 'ch3ch3chch3'],
                   ['tbu', 'cch3ch3ch3', 'c(ch3)3', 'ch3ch3ch3c', '(ch3)3c'],
                   ['vi', 'chch2', 'c2h3', 'h2chc', 'h3c2'],
                   ['ey', 'cch', 'c2h', 'hcc', 'hc2'],
                   ['ph', 'c6h5', 'h5c6'],
                   ['am', 'nh2', 'h2n'],
                   ['im', 'chnh', 'cnh2', 'nhch'],
                   ['cn', 'nc'],
                   ['oh', 'ho'],
                   ['ome', 'meo', 'och3', 'ch3o'],
                   ['al', 'cho', 'coh', 'och', 'ohc'],
                   ['ac', 'coch3', 'cch3o'],
                   ['ca', 'cooh', 'co2h', 'hooc', 'ho2c'],
                   ['nt', 'no2', 'o2n'],
                   ['f'],
                   ['tfm', 'cf3', 'f3c'],
                   ['sh', 'hs'],
                   ['sf', 'so2h', 'sooh', 'sho2', 'ho2s', 'hso2'],
                   ['ms', 'sfme', 'mesf', 'sfch3', 'so2me', 'so2ch3'],
                   ['cl']]
        for subl in synlist:
            for item in subl:
                self.syn[item] = subl[0]

    def _populate_elem(self):
        """Adds element labels to self.elem."""
        self.elem['h'] = np.array(['H'])
        self.elem['d'] = np.array(['D'])
        self.elem['me'] = np.array(['C', 'H', 'H', 'H'])
        self.elem['vi'] = np.array(['C', 'C', 'H', 'H', 'H'])
        self.elem['ey'] = np.array(['C', 'C', 'H'])
        self.elem['ph'] = np.array(['C', 'C', 'C', 'C', 'C', 'C',
                                    'H', 'H', 'H', 'H', 'H'])
        self.elem['am'] = np.array(['N', 'H', 'H'])
        self.elem['im'] = np.array(['C', 'N', 'H', 'H'])
        self.elem['cn'] = np.array(['C', 'N'])
        self.elem['oh'] = np.array(['O', 'H'])
        self.elem['al'] = np.array(['C', 'O', 'H'])
        self.elem['nt'] = np.array(['N', 'O', 'O'])
        self.elem['f'] = np.array(['F'])
        self.elem['sh'] = np.array(['S', 'H'])
        self.elem['sf'] = np.array(['S', 'O', 'O', 'H'])
        self.elem['cl'] = np.array(['Cl'])

    def _populate_xyz(self):
        """Adds cartesian geometries to self.xyz.

        For all substituents, the bonding atom is at the origin,
        the bonding axis is the z-axis and the plane axis
        is the y-axis.
        """
        self.xyz['h'] = np.array([[ 0.000,  0.000,  0.000]])
        self.xyz['d'] = np.array([[ 0.000,  0.000,  0.000]])
        self.xyz['me'] = np.array([[ 0.000,  0.000,  0.000],
                                   [ 0.511, -0.886,  0.377],
                                   [ 0.511,  0.886,  0.377],
                                   [-1.023,  0.000,  0.377]])
        self.xyz['vi'] = np.array([[ 0.000,  0.000,  0.000],
                                   [-1.124,  0.000,  0.730],
                                   [ 0.971,  0.000,  0.495],
                                   [-1.067,  0.000,  1.818],
                                   [-2.095,  0.000,  0.235]])
        self.xyz['ey'] = np.array([[ 0.000,  0.000,  0.000],
                                   [ 0.000,  0.000,  1.210],
                                   [ 0.000,  0.000,  2.280]])
        self.xyz['ph'] = np.array([[ 0.000,  0.000,  0.000],
                                   [-1.212,  0.000,  0.700],
                                   [ 1.212,  0.000,  0.700],
                                   [-1.212,  0.000,  2.100],
                                   [ 1.212,  0.000,  2.100],
                                   [ 0.000,  0.000,  2.800],
                                   [-2.156,  0.000,  0.155],
                                   [ 2.156,  0.000,  0.155],
                                   [-2.156,  0.000,  2.645],
                                   [ 2.156,  0.000,  2.645],
                                   [ 0.000,  0.000,  3.890]])
        self.xyz['am'] = np.array([[ 0.000,  0.000,  0.000],
                                   [-0.577, -0.771,  0.332],
                                   [-0.577,  0.771,  0.332]])
        self.xyz['im'] = np.array([[ 0.000,  0.000,  0.000],
                                   [-1.082,  0.000,  0.703],
                                   [ 0.980,  0.000,  0.499],
                                   [-0.869,  0.000,  1.710]])
        self.xyz['cn'] = np.array([[ 0.000,  0.000,  0.000],
                                   [ 0.000,  0.000,  1.136]])
        self.xyz['oh'] = np.array([[ 0.000,  0.000,  0.000],
                                   [-0.913,  0.000,  0.297]])
        self.xyz['al'] = np.array([[ 0.000,  0.000,  0.000],
                                   [-1.011,  0.000,  0.700],
                                   [ 0.998,  0.000,  0.463]])
        self.xyz['nt'] = np.array([[ 0.000,  0.000,  0.000],
                                   [-1.105,  0.000,  0.563],
                                   [ 1.105,  0.000,  0.563]])
        self.xyz['f'] = np.array([[ 0.000,  0.000,  0.000]])
        self.xyz['sh'] = np.array([[ 0.000,  0.000,  0.000],
                                   [-1.331,  0.000,  0.156]])
        self.xyz['sf'] = np.array([[ 0.000,  0.000,  0.000],
                                   [ 0.548, -1.266,  0.448],
                                   [ 0.548,  1.266,  0.448],
                                   [-1.311,  0.000,  0.279]])
        self.xyz['cl'] = np.array([[ 0.000,  0.000,  0.000]])

    def _add_comb(self):
        """Adds substituents made by combining multiple substituents."""
        self.elem['et'], self.xyz['et'] = self.add_subs('me', 'me')
        self.elem['npr'], self.xyz['npr'] = self.add_subs('me', 'me', 'me')
        self.elem['ipr'], self.xyz['ipr'] = self.add_subs('me', 'me', 'me',
                                                          inds=1)
        self.elem['nbu'], self.xyz['nbu'] = self.add_subs('me', 'me', 'me',
                                                          'me')
        self.elem['ibu'], self.xyz['ibu'] = self.add_subs('me', 'me', 'me',
                                                          'me', inds=[2, 1, -1])
        self.elem['tbu'], self.xyz['tbu'] = self.add_subs('me', 'me', 'me',
                                                          'me', inds=1)
        self.elem['ome'], self.xyz['ome'] = self.add_subs('oh', 'me')
        self.elem['ac'], self.xyz['ac'] = self.add_subs('al', 'me')
        self.elem['ca'], self.xyz['ca'] = self.add_subs('al', 'oh')
        self.elem['tfm'], self.xyz['tfm'] = self.add_subs('me', 'f', 'f', 'f',
                                                          inds=1)
        self.elem['ms'], self.xyz['ms'] = self.add_subs('sf', 'me')

    def get_sub(self, label):
        """Returns the element list and cartesian geometry of a
        substituent.

        Parameters
        ----------
        label : str
            The substituent label of the desired substituent.

        Returns
        -------
        elem : (N,) ndarray
            The atomic symbols of the substituent.
        xyz : (N, 3) ndarray
            The atomic cartesian coordinates of the substituent.
        """
        lbl = self.syn[label.lower()]
        return self.elem[lbl], self.xyz[lbl]

    def add_subs(self, *lbls, inds=-1):
        """Returns the element list and cartesian geometry from a
        combination of substituents.

        Parameters
        ----------
        lbls : list
            A list of substituent labels to be combined.
        inds : int or array_like, optional
            The indices for substitution between substituents. Setting
            inds=-1 (default) makes the last atom the subtituted atom.
            Otherwise a list of indices can be given for the first of
            each pair of substituents.

        Returns
        -------
        elem : (N,) ndarray
            The atomic symbols of the combined substituent.
        xyz : (N, 3) ndarray
            The atomic cartesian coordinates of the combined substituent.
        """
        if isinstance(inds, int):
            inds = (len(lbls) - 1) * [inds]
        elif len(inds) != len(lbls) - 1:
            raise ValueError('Number of inds != number of labels - 1')

        rot = 0
        lbl0 = self.syn[lbls[0].lower()]
        elem = self.elem[lbl0]
        xyz = self.xyz[lbl0]
        for i, label in zip(inds, lbls[1:]):
            dist = np.linalg.norm(xyz - xyz[i], axis=1)
            dist[i] += np.max(dist)
            ibond = np.argmin(dist)
            rot = (rot + 1) % 2
            ax = con.unit_vec(xyz[i] - xyz[ibond])
            lbl = self.syn[label.lower()]
            new_elem = self.elem[lbl]
            new_xyz = displace.rotate(self.xyz[lbl], rot*np.pi, 'Z')
            new_xyz = displace.align_axis(new_xyz, 'Z', ax)
            blen = con.get_covrad(elem[ibond]) + con.get_covrad(new_elem[0])
            new_xyz += xyz[ibond] + blen * ax
            elem = np.hstack((np.delete(elem, i), new_elem))
            xyz = np.vstack((np.delete(xyz, i, axis=0), new_xyz))

        return elem, xyz


def import_sub(label):
    """Returns the element list and cartesian geometry of a substituent
    given its label.

    Parameters
    ----------
    label : str
        The substituent label.

    Returns
    -------
    elem : (N,) ndarray
        The atomic symbols of the substituent.
    xyz : (N, 3) ndarray
        The atomic cartesian coordinates of the substituent.
    """
    lib = SubLib()
    return lib.get_sub(label)


def subst(elem, xyz, sublbl, isub, ibond=None, pl=None, vec=None):
    """Returns a molecular geometry with an specified atom replaced by
    substituent.

    Labels are case-insensitive. The index isub gives the position to be
    substituted. If specified, ibond gives the atom bonded to the
    substituent. Otherwise, the nearest atom to isub is used. The
    orientation of the substituent can be given as a vector (the plane
    normal) or an index (the plane containing isub, ibond and pl).

    If isub is given as a list, the entire list of atoms is removed
    and the first index is treated as the position of the substituent.

    Parameters
    ----------
    elem : (N,) array_like
        The atomic symbols of the unsubstituted molecule.
    xyz : (N, 3) array_like
        The atomic cartesian coordinates of the unsubstituted molecule.
    sublbl : str
        The substituent label.
    isub : int or list
        The atomic index (or indices) to be replaced by the substituent.
    ibond : int, optional
        The atomic index of the atom bonded to position isub. If None
        (default), the nearest atom is chosen.
    pl : int or array_like, optional
        The atomic index or vector defining the xz-plane of the
        substituent. If an index is given, the plane normal to the
        isub-ibond-pl plane is used. If None (default), the plane
        is arbitrarily set to [1, 1, 1] and the bond axis is projected
        out.
    vec : (N, 3) array_like, optional
        The atomic cartesian vectors of the unsubstitued molecule. Default
        is None.

    Returns
    -------
    new_elem : (N,) ndarray
        The atomic symbols of the substituted molecule.
    new_xyz : (N, 3) ndarray
        The atomic cartesian coordinates of the substituted molecule.
    new_vec : (N, 3) ndarray
        The atomic cartesian vectors of the substituted molecule.
        Substituent atoms are all set of zero. If vec is None, new_vec
        is all zeros.
    """
    elem = np.array(elem)
    xyz = np.atleast_2d(xyz)
    if not isinstance(isub, int):
        ipos = isub[0]
    else:
        isub = [isub]
        ipos = isub[0]

    if ibond is None:
        dist = np.linalg.norm(xyz - xyz[ipos], axis=1)
        dist[ipos] += np.max(dist)
        ibond = np.argmin(dist)
    elif ibond == ipos:
        raise ValueError('sub and bond indices cannot be the same')

    ax = con.unit_vec(xyz[ipos] - xyz[ibond])
    if pl is None:
        # choose an arbitrary axis and project out the bond axis
        pl = np.ones(3)
        pl -= np.dot(pl, ax) * ax
    elif isinstance(pl, int):
        if pl == ipos:
            raise ValueError('plane and sub indices cannot be the same')
        elif pl == ibond:
            raise ValueError('plane and bond indices cannot be the same')
        pl = np.cross(xyz[ipos] - xyz[ibond], xyz[pl] - xyz[ibond])

    sub_el, sub_xyz = import_sub(sublbl)
    if elem[ipos] == sub_el[0]:
        blen = np.linalg.norm(xyz[ipos] - xyz[ibond])
    else:
        blen = con.get_covrad(elem[ibond]) + con.get_covrad(sub_el[0])

    # rotate to correct orientation and displace to correct position
    sub_xyz = displace.align_axis(sub_xyz, 'Z', ax)
    sub_pl = displace.align_axis([0., 1., 0.], 'Z', ax)
    sub_xyz = displace.align_axis(sub_xyz, sub_pl, pl)
    sub_xyz += xyz[ibond] + blen * ax

    # build the final geometry
    ind1 = [i for i in range(ipos) if i not in isub[1:]]
    ind2 = [i for i in range(ipos+1, len(elem)) if i not in isub[1:]]
    new_elem = np.hstack((elem[ind1], sub_el, elem[ind2]))
    new_xyz = np.vstack((xyz[ind1], sub_xyz, xyz[ind2]))
    if vec is None:
        return new_elem, new_xyz, None
    else:
        new_vec = np.vstack((vec[ind1], np.zeros((len(sub_el), 3)), vec[ind2]))
        return new_elem, new_xyz, new_vec



"""
The Molecule and MoleculeBundle objects and tools for generating and
querying molecular geometries.

Molecule creates a saved copy of the geometry after input for reversion
after an operation. Can add/remove individual atoms or groups or set the
full geometry.

Likewise, MoleculeBundle creates a saved copy of a set molecular
geometries. Input files with multiple geometries can be read to a bundle.
"""
import numpy as np
import gimbal.fileio as fileio
import gimbal.displace as displace
import gimbal.measure as measure
import gimbal.substitute as substitute
import gimbal.constants as con
import gimbal.kabsch as kabsch


class BaseMolecule(object):
    """
    Basic object containing molecular geometry and functions for setting and
    changing the geometry.

    All methods of BaseMolecule involve setting and saving the molecular
    geometry. There are no dependencies to other modules.

    Attributes
    ----------
    elem : (N,) ndarray
        The atomic symbols.
    xyz : (N, 3) ndarray
        The atomic cartesian coordinates.
    vec : (N, 3) ndarray
        The atomic cartesian vectors.
    comment : str
        The comment line.
    natm : int
        The number of atoms N. Automatically set during initialization.
    print_vec : bool
        Specifies if the vectors should be printed/written. Automatically
        set to False if vec is None (default).
    save_elem : (N,) ndarray
        The saved atomic symbols.
    save_xyz : (N, 3) ndarray
        The saved atomic cartesian coordinates.
    save_vec : (N, 3) ndarray
        The saved atomic cartesian vectors.
    save_comment : str
        The saved comment line.
    save_print_vec : bool
        Specifies if the vectors should be printed/written. Automatically
        set to False if vec is None (default).
    saved : bool
        Specifies if the ``save_...`` values are up to date.
    """
    def __init__(self, elem=[], xyz=np.empty((0, 3)), vec=None, comment=''):
        self._elem = np.atleast_1d(np.array(elem, dtype=str))
        self._xyz = np.atleast_2d(np.array(xyz, dtype=float))
        self._comment = repr(str(comment))[1:-1]
        if vec is None:
            self._vec = np.zeros_like(self._xyz)
            self.print_vec = False
        else:
            self._vec = np.atleast_2d(np.array(vec, dtype=float))
            self.print_vec = True
        self.save()

    def __repr__(self):
        nelem = len(self.elem)
        fmt = '[{:>2s},' + 2*'{:16.8e},' + '{:16.8e}'
        if self.print_vec:
            xyzvec = np.hstack((self._xyz, self._vec))
            fmt += ',\n' + 6*' ' + 2*'{:16.8e},' + '{:16.8e}]'
        else:
            xyzvec = self._xyz
            fmt += ']'

        if nelem == 0:
            fstr = 'BaseMolecule({!r}, ['.format(self._comment)
        else:
            fstr = 'BaseMolecule({!r},\n ['.format(self._comment)

        if nelem > 10:
            fstr += fmt.format(self._elem[0], *xyzvec[0])
            for i in range(1, 3):
                fstr += ',\n  ' + fmt.format(self._elem[i], *xyzvec[i])
            fstr += ',\n  ...'
            for i in range(-3, 0):
                fstr += ',\n  ' + fmt.format(self._elem[i], *xyzvec[i])
        else:
            for i in range(nelem):
                if i != 0:
                    fstr += ',\n  '
                fstr += fmt.format(self._elem[i], *xyzvec[i])

        fstr += '])'
        return fstr

    def __str__(self):
        nelem = len(self.elem)
        fmt = '[{:>2s}' + 3*'{:14.8f}'
        if self.print_vec:
            xyzvec = np.hstack((self._xyz, self._vec))
            fmt += '\n' + 4*' ' + 3*'{:14.8f}' + ']'
        else:
            xyzvec = self._xyz
            fmt += ']'
        fstr = ''
        if self.comment != '':
            fstr += '{!s}\n['.format(self.comment)
        else:
            fstr += '['
        if nelem > 10:
            fstr += fmt.format(self._elem[0], *xyzvec[0])
            for i in range(1, 3):
                fstr += '\n ' + fmt.format(self._elem[i], *xyzvec[i])
            fstr += '\n ...'
            for i in range(-3, 0):
                fstr += '\n ' + fmt.format(self._elem[i], *xyzvec[i])
        else:
            for i in range(nelem):
                if i != 0:
                    fstr += '\n '
                fstr += fmt.format(self._elem[i], *xyzvec[i])
        fstr += ']'
        return fstr

    def _check(self, ielem=None, ixyz=None, ivec=None):
        """Checks that xyz is 3D and len(elem) = len(xyz).

        Parameters
        ----------
        ielem : (N,) array_like, optional
            The new atomic symbols to be checked. If None (default), the
            values of self._elem are checked instead.
        ixyz : (N, 3) array_like, optional
            The new atomic cartesian coordiantes to be checked. If None
            (default), the values of self._xyz are checked instead.
        ivec : (N, 3) array_like, optional
            The new atomic cartesian vectors to be checked. If None
            (default), the values of self._vec are checked instead.

        Raises
        ------
        ValueError
            When the shapes of ixyz, elem and vec don't match or ixyz
            and ivec aren't 3D.
        """
        if ielem is None:
            ielem = self._elem
        if ixyz is None:
            ixyz = self._xyz
        if ivec is None:
            ivec = self._vec
        if ixyz.shape[1] != 3:
            raise ValueError('Molecular geometry must be 3-dimensional.')
        if ivec.shape[1] != 3:
            raise ValueError('Molecular vector must be 3-dimensional.')
        natm = len(ielem)
        if natm != len(ixyz):
            raise ValueError('Number of element labels not equal '
                             'to number of cartesian vectors.')
        elif ixyz.shape != ivec.shape:
            if self.print_vec:
                raise ValueError('Cartesian geometry and vector must have '
                                 'the same number of elements.')
            else:
                self._vec = np.zeros_like(ixyz)
        self.natm = natm

    @property
    def elem(self):
        """Gets the value of elem."""
        self._check()
        return self._elem

    @elem.setter
    def elem(self, val):
        """Sets the value of elem."""
        val = np.atleast_1d(np.array(val, dtype=str))
        self._check(ielem=val)
        self._elem = val
        self.saved = False

    @property
    def xyz(self):
        """Gets the value of xyz."""
        self._check()
        return self._xyz

    @xyz.setter
    def xyz(self, val):
        """Sets the value of xyz."""
        val = np.atleast_2d(np.array(val, dtype=float))
        self._check(ixyz=val)
        self._xyz = val
        self.saved = False

    @property
    def vec(self):
        """Gets the value of vec."""
        self._check()
        return self._vec

    @vec.setter
    def vec(self, val):
        """Sets the value of vec."""
        val = np.atleast_2d(np.array(val, dtype=float))
        self._check(ivec=val)
        self._vec = val
        self.print_vec = True
        self.saved = False

    @property
    def comment(self):
        """Gets the value of comment."""
        return self._comment

    @comment.setter
    def comment(self, val):
        """Sets the value of comment."""
        self._comment = repr(str(val))[1:-1]
        self.saved = False

    def copy(self, comment=None):
        """Creates a copy of the BaseMolecule object.

        Parameters
        ----------
        comment : str
            A comment line for the copied BaseMolecule. If None (default),
            self._comment is used.

        Returns
        -------
        BaseMolecule
            A copy of the BaseMolecule instance.
        """
        self._check()
        if comment is None:
            comment = self._comment
        return BaseMolecule(np.copy(self._elem), np.copy(self._xyz),
                            np.copy(self._vec), comment)

    def save(self):
        """Saves molecular properties to ``save_`` variables."""
        self._check()
        self.save_elem = np.copy(self._elem)
        self.save_xyz = np.copy(self._xyz)
        self.save_vec = np.copy(self._vec)
        self.save_comment = np.copy(self._comment)
        self.save_print_vec = np.copy(self.print_vec)
        self.saved = True

    def revert(self):
        """Reverts properties to ``save_`` variables."""
        if not self.saved:
            self._elem = np.copy(self.save_elem)
            self._xyz = np.copy(self.save_xyz)
            self._vec = np.copy(self.save_vec)
            self._comment = np.copy(self.save_comment)
            self.print_vec = np.copy(self.save_print_vec)
        self.saved = True

    def add_atoms(self, new_elem, new_xyz, new_vec=None):
        """Adds atoms(s) to molecule.

        Parameters
        ----------
        new_elem : (n,) array_like
            The new atomic symbols to be added to the current symbol
            list.
        new_xyz : (n, 3) array_like
            The new atomic cartesian coordinates to be added to the
            current coordinate list.
        new_vec : (n, 3) array_like, optional
            The new atomic cartesian vectors to be added to the current
            vector list. If None (default), zeros are added to match
            new_xyz.
        """
        new_xyz = np.atleast_2d(np.array(new_xyz, dtype=float))
        elems = np.hstack((self._elem, new_elem))
        xyzs = np.vstack((self._xyz, new_xyz))
        if new_vec is not None:
            new_vec = np.atleast_2d(np.array(new_vec, dtype=float))
            vecs = np.vstack((self._vec, new_vec))
        else:
            vecs = None
        self._check(ielem=elems, ixyz=xyzs, ivec=vecs)
        self._elem = elems
        self._xyz = xyzs
        if new_vec is not None:
            self._vec = vecs
            self.print_vec = True
        self.saved = False

    def rm_atoms(self, ind):
        """Removes atom(s) from molecule by index.

        Parameters
        ----------
        ind : int or array_like
            Index or indices of atoms to be removed.
        """
        self._elem = np.delete(self._elem, ind)
        self._xyz = np.delete(self._xyz, ind, axis=0)
        self._vec = np.delete(self._vec, ind, axis=0)
        self._check()
        self.saved = False

    def rearrange(self, new_ind, old_ind=None):
        """Moves atom(s) from old_ind to new_ind.

        Parameters
        ----------
        new_ind : array_like
            New indices for the atoms.
        old_ind : array_like, optional
            Old indices for the atoms. If None (default), new_ind must
            include all atoms.
        """
        if old_ind is None:
            old_ind = range(self.natm)
        new, old = _rearrange_inds(new_ind, old_ind)
        self._xyz[old] = self._xyz[new]
        self._vec[old] = self._vec[new]
        self._elem[old] = self._elem[new]
        self._check()
        self.saved = False


class Molecule(BaseMolecule):
    """
    More advanced Molecule object which inherits from BaseMolecule.

    A dummy atom XM is added to index 0 at the centre of mass of the
    molecule. Indices thus match the regular definition. Accessor methods
    get_elem and get_xyz return the geometry without the index 0 dummy atom.

    Molecule can also read from and write to input files given by a filename
    or an open file object. Other methods are derived from different modules.

    Attributes
    ----------
    elem : (N+1,) ndarray
        The atomic symbols.
    xyz : (N+1, 3) ndarray
        The atomic cartesian coordinates.
    vec : (N+1, 3) ndarray
        The atomic cartesian vectors.
    comment : str
        The comment line.
    natm : int
        The number of atoms N. Automatically set during initialization.
    print_vec : bool
        Specifies if the vectors should be printed/written. Automatically
        set to False if vec is None (default).
    """
    def __repr__(self):
        baserepr = BaseMolecule.__repr__(self)
        return baserepr.replace('BaseMolecule', 'Molecule', 1)

    def __str__(self):
        base = BaseMolecule(self.elem[1:], self.xyz[1:],
                            self.vec[1:] if self.print_vec else None,
                            self.comment)
        return base.__str__()

    def __add__(self, other):
        return MoleculeBundle(_add_type(self, other))

    def _add_centre(self):
        """Adds a dummy atom at index 0 at molecular centre of mass."""
        pos = displace.get_centremass(self._elem, self._xyz)
        if self._elem[0] == 'XM':
            self._xyz[0] = pos
        else:
            self._elem = np.hstack(('XM', self._elem))
            self._xyz = np.vstack((pos, self._xyz))
            self._vec = np.vstack(([0, 0, 0], self._vec))

    def _check(self, ielem=None, ixyz=None, ivec=None):
        """Checks that xyz is 3D and len(elem) = len(xyz) and that dummy
        atom is at centre of mass.

        Parameters
        ----------
        ielem : (N,) array_like, optional
            The new atomic symbols to be checked. If None (default), the
            values of self._elem are checked instead.
        ixyz : (N, 3) array_like, optional
            The new atomic cartesian coordiantes to be checked. If None
            (default), the values of self._xyz are checked instead.
        ivec : (N, 3) array_like, optional
            The new atomic cartesian vectors to be checked. If None
            (default), the values of self._vec are checked instead.
        """
        BaseMolecule._check(self, ielem=ielem, ixyz=ixyz, ivec=ivec)
        if self.natm > 0:
            self._add_centre()
            self.natm = len(self._elem) - 1

    def copy(self, comment=None):
        """Creates a copy of the Molecule object.

        Parameters
        ----------
        comment : str
            A comment line for the copied Molecule. If None (default),
            self._comment is used.

        Returns
        -------
        Molecule
            A copy of the Molecule instance.
        """
        self._check()
        vec = self._vec if self.print_vec else None
        if comment is None:
            comment = self._comment
        return Molecule(np.copy(self._elem), np.copy(self._xyz),
                        vec, comment)

    def rearrange(self, new_ind, old_ind=None):
        """Moves atom(s) from old_ind to new_ind.

        Parameters
        ----------
        new_ind : array_like
            New indices for the atoms.
        old_ind : array_like, optional
            Old indices for the atoms. If None (default), new_ind must
            include all atoms except for the dummy atom (index 0).
        """
        if old_ind is None:
            old_ind = range(1, self.natm+1)
        BaseMolecule.rearrange(self, new_ind, old_ind)

    def read(self, infile, fmt='auto', **kwargs):
        """Reads single geometry from input file in provided format.

        Parameters
        ----------
        infile : file or str
            The open input file or filename.
        fmt : str, optional
            The file format. Default is auto (i.e. :func:`fileio.read_auto`).
        kwargs : dict, optional
            Additional keyword arguments for the read function.
        """
        (self._elem, self._xyz, ivec,
         self._comment) = fileio.read_single(infile, fmt=fmt, **kwargs)
        if ivec is None:
            self._vec = np.zeros_like(self._xyz)
        else:
            self._vec = ivec

        self._check()
        self.saved = False

    def write(self, outfile, fmt='auto', **kwargs):
        """Writes geometry to an output file in provided format.

        Parameters
        ----------
        outfile : file or str
            The open output file or filename.
        fmt : str, optional
            The file format. Default is auto (i.e. :func:`fileio.write_auto`).
        kwargs : dict, optional
            Additional keyword arguments for the write function.
        """
        vec = self._vec[1:] if self.print_vec else None
        moldat = (self._elem[1:], self._xyz[1:], vec, self._comment)
        fileio.write_single(outfile, moldat, fmt=fmt, **kwargs)

    def get_mass(self):
        """Returns atomic masses.

        Returns
        -------
        ndarray
            The array of atomic masses.
        """
        return con.get_mass(self._elem)

    def get_formula(self):
        """Gets the atomic formula based on the element list.

        Returns
        -------
        str
            The molecular atomic formula as a string.
        """
        elem = [sym for sym in self.elem if 'X' not in sym]
        atm, num = np.unique(elem, return_counts=True)
        fstr = ''
        for a, n in zip(atm, num):
            if n == 1:
                fstr += a
            else:
                fstr += a + str(n)

        return fstr

    def measure(self, coord, *inds, **kwargs):
        """Returns a coordinate based on its indices in the molecule.

        Parameters
        ----------
        coord : str
            The coordinate specification, given by function names
            in :mod:`measure`.
        inds : list
            The indices used in the coordinate function.
        kwargs : dict, optional
            Additional keyword arguments for the measure functions.

        Returns
        -------
        float
            The value of the specified coordinate.
        """
        self._check()
        coord_func = getattr(measure, coord)
        return coord_func(self.xyz, *inds, **kwargs)

    def centre_mass(self):
        """Places the centre of mass at the origin."""
        self._check()
        self._xyz -= self._xyz[0]
        self.saved = False

    def translate(self, amp, axis, ind=None, units='ang'):
        """Translates the molecule along a given axis.

        Momenta are difference vectors and are not translated.

        Parameters
        ----------
        amp : float
            The distance for translation.
        axis : array_like or str
            The axis of translation, parsed by :func:`displace._parse_axis`.
        ind : array_like, optional
            List of atomic indices to specify which atoms are displaced.
            If ind is None (default) then all atoms are displaced.
        units : str, optional
            The units of length for displacement. Default is angstroms.
        """
        self.xyz = displace.translate(self._xyz, amp, axis, ind=ind,
                                      units=units)
        self.saved = False

    def rotate(self, amp, axis, ind=None, origin=np.zeros(3), det=1,
               units='rad'):
        """Rotates the molecule about a given axis from a given origin.

        If vectors are non-zero, they will be rotated about the
        same origin. Reflections and improper rotations can be done
        by setting det=-1.

        Parameters
        ----------
        amp : float
            The angle of rotation.
        axis : array_like or str
            The axis of rotation, parsed by :func:`displace._parse_axis`.
        ind : array_like, optional
            List of atomic indices to specify which atoms are displaced.
            If ind is None (default) then all atoms are displaced.
        origin : (3,) array_like, optional
            The origin of rotation. Default is the cartesian origin.
        det : float, optional
            The determinant of the rotation. 1 (default) is a proper rotation
            and -1 is an improper rotation (rotation + reflection).
        units : str, optional
            The units of length for displacement. Default is angstroms.
        """
        kwargs = dict(ind=ind, origin=origin, det=det, units=units)
        self.xyz = displace.rotate(self._xyz, amp, axis, **kwargs)
        if self.print_vec:
            self.vec = displace.rotate(self._vec, amp, axis, **kwargs)
        self.saved = False

    def match_to_ref(self, ref_bundle, **kwargs):
        """Tests the molecule against a set of references in a bundle.

        At the moment, vectors are not properly rotated.

        Parameters
        ----------
        ref_bundle : MoleculeBundle
            A molecule bundle object containing the set of reference
            geometries.
        kwargs : dict
            Additional keyword arguments used in :func:`kabsch.opt_permute`.

        Returns
        -------
        int
            The index of the optimal ref geometry.
        """
        vec = self._vec if self.print_vec else None
        reflist = [mol._xyz for mol in ref_bundle.molecules]
        self.xyz, ind = kabsch.opt_ref(self._elem, self._xyz, reflist, **kwargs)
        return ind

    def subst(self, lbl, isub, ibond=None, pl=None):
        """Replaces an atom or set of atoms with a substituent.

        Parameters
        ----------
        lbl : str
            The substituent label.
        isub : int or list
            The atomic index (or indices) to be replaced by the substituent.
        ibond : int, optional
            The atomic index of the atom bonded to position isub. If None
            (default), the nearest atom is chosen.
        pl : int or array_like, optional
            The atomic index or vector defining the xz-plane of the
            substituent. See :func:`substitute.subst` for more details.
        """
        args = (self._elem, self._xyz, lbl, isub)
        kwargs = dict(ibond=ibond, pl=pl, vec=self._vec)
        self._elem, self._xyz, vec = substitute.subst(*args, **kwargs)
        if self.print_vec:
            self._vec = vec
        else:
            self._vec = np.zeros_like(self._xyz)
        self._check()


class MoleculeBundle(object):
    """
    Object containing a set of molecules in the form of Molecule
    objects.

    Attributes
    ----------
    molecules : ndarray
        The array of Molecule objects.
    nmol : int
        The number of molecules, set automatically during initialization.
    save_molecules : ndarray
        The saved array of Molecule objects.
    saved : bool
        Specifies if the save_molecules value is up to date.
    """
    def __init__(self, molecules=[]):
        self._molecules = np.atleast_1d(np.array(molecules, dtype=object,
                                                 copy=False))
        self._check()
        self.save()

    def __repr__(self):
        nmol = len(self.molecules)
        if nmol == 0:
            fstr = ''
        else:
            fstr = '\n '
            if nmol > 6:
                fstr += self._join_str(',\n\n ', self._molecules[:2], 'r')
                fstr += ',\n\n ...,\n\n '
                fstr += self._join_str(',\n\n ', self._molecules[-2:], 'r')
            else:
                fstr += self._join_str(',\n\n ', self._molecules, 'r')
        return 'MoleculeBundle({:s})'.format(fstr)

    def __str__(self):
        nmol = len(self.molecules)
        if nmol == 0:
            fstr = ''
        elif nmol > 6:
            fstr = self._join_str('\n\n ', self._molecules[:2], 's')
            fstr += '\n\n ...\n\n '
            fstr += self._join_str('\n\n ', self._molecules[-2:], 's')
        else:
            fstr = self._join_str('\n\n ', self._molecules, 's')
        return '[{:s}]'.format(fstr)

    @staticmethod
    def _join_str(jn, lst, typ):
        """Returns a string of list elements joined by a separator with
        newline characters added."""
        fmt = '{!' + typ + '}'
        return jn.join(fmt.format(l, t=typ).replace('\n', '\n ')
                       for l in lst)

    def __add__(self, other):
        return MoleculeBundle(_add_type(self, other))

    def __iadd__(self, other):
        return MoleculeBundle(_add_type(self, other))

    def _check(self, imol=None):
        """Check that all bundle objects are Molecule type.

        Parameters
        ----------
        imol : ndarray, optional
            The new array of Molecule objects to be checked. If None
            (default), self._molecules is checked instead.

        Raises
        ------
        TypeError
            When the type of elements of imol aren't Molecule.
        """
        if imol is None:
            imol = self._molecules
        for mol in imol:
            if not isinstance(mol, Molecule):
                raise TypeError('Elements of molecule bundle must be '
                                'Molecule type.')
        self.nmol = len(self._molecules)

    @property
    def molecules(self):
        """Gets the value of molecules."""
        self._check()
        return self._molecules

    @molecules.setter
    def molecules(self, val):
        """Sets the value of molecules."""
        val = np.atleast_1d(np.array(val, dtype=object, copy=False))
        self._check(imol=val)
        self._molecules = val
        self.saved = False

    def copy(self):
        """Returns a copy of the MoleculeBundle object.

        Returns
        -------
        MoleculeBundle
            A copy of the MoleculeBundle instance.
        """
        return MoleculeBundle([mol.copy() for mol in self._molecules])

    def save(self):
        """Saves all molecules in the bundle and sets save_molecules."""
        self._check()
        for mol in self._molecules:
            mol.save()
        self.save_molecules = np.copy(self._molecules)
        self.saved = True

    def revert(self):
        """Reverts each molecule in the bundle to the saved variables."""
        if not self.saved:
            self._molecules = np.copy(self.save_molecules)
        for mol in self._molecules:
            mol.revert()
        self.saved = True

    def rearrange(self, new_ind, old_ind=None):
        """Moves molecule(s) from old_ind to new_ind in bundle.

        Parameters
        ----------
        new_ind : array_like
            New indices for the molecules.
        old_ind : array_like, optional
            Old indices for the molecules. If None (default), new_ind must
            include all molecules.
        """
        if old_ind is None:
            old_ind = range(self.nmol)
        new, old = _rearrange_inds(new_ind, old_ind)
        self._molecules[old] = self._molecules[new]
        self.saved = False

    def add_molecules(self, new_molecules):
        """Adds molecule(s) to the bundle.

        Parameters
        ----------
        new_molecules : (n,) array_like
            The new Molecule objects to be added to the current molecule
            list.
        """
        mols = np.hstack((self._molecules, new_molecules))
        self._check(imol=mols)
        self._molecules = mols
        self.saved = False

    def rm_molecules(self, ind):
        """Removes molecule(s) from the bundle by index.

        Parameters
        ----------
        ind : int or array_like
            Index or indices of molecules to be removed.
        """
        self._molecules = np.delete(self._molecules, ind)
        self._check()
        self.saved = False

    def read(self, inflist, fmt='auto', **kwargs):
        """Reads all geometries from input file in provided format.

        Parameters
        ----------
        infile : array_like
            The open input files or filenames.
        fmt : str, optional
            The file format. Default is auto (i.e. :func:`fileio.read_auto`).
        kwargs : dict, optional
            Additional keyword arguments for the read function.
        """
        moldat = fileio.read_multiple(inflist, fmt=fmt, **kwargs)
        new_mol = [Molecule(*dat) for dat in moldat]
        self._molecules = np.hstack((self.molecules, new_mol))
        self._check()
        self.saved = False

    def write(self, outfile, fmt='auto', **kwargs):
        """Writes geometries to an output file in provided format.

        Parameters
        ----------
        outfile : file or str
            The open output file or filename.
        fmt : str, optional
            The file format. Default is auto (i.e. :func:`fileio.write_auto`).
        kwargs : dict, optional
            Additional keyword arguments for the write function.
        """
        moldat = np.empty(self.nmol, dtype=object)
        for i, mol in enumerate(self.molecules):
            if mol.print_vec:
                moldat[i] = (mol.elem[1:], mol.xyz[1:], mol.vec[1:],
                             mol.comment)
            else:
                moldat[i] = (mol.elem[1:], mol.xyz[1:], None, mol.comment)

        fileio.write_multiple(outfile, moldat, fmt=fmt, **kwargs)

    def measure(self, coord, *inds, **kwargs):
        """Returns a list of coordinates based on index in molecules.

        Parameters
        ----------
        coord : str
            The coordinate specification, given by function names
            in :mod:`measure`.
        inds : list
            The indices used in the coordinate function.
        kwargs : dict, optional
            Additional keyword arguments for the measure functions.

        Returns
        -------
        ndarray
            The values of the specified coordinate for each molecule.
        """
        return np.array([mol.measure(coord, *inds, **kwargs)
                         for mol in self.molecules])

    def match_to_ref(self, ref_bundle, plist=None, equiv=None, wgt=None,
                     ind=None, cent=None):
        """Tests the molecules in the current bundle against
        a set of references in another bundle.

        Returns a set of bundles corresponding to the reference indices.

        Parameters
        ----------
        ref_bundle : MoleculeBundle
            A molecule bundle object containing the set of reference
            geometries.
        kwargs : dict
            Additional keyword arguments used in :func:`kabsch.opt_permute`.

        Returns
        -------
        ndarray
            The indices of the optimal ref geometries for each molecule.
        """
        kwargs = dict(plist=plist, equiv=equiv, wgt=wgt, ind=ind, cent=cent)
        inds = np.empty(self.nmol)
        for i, mol in enumerate(self._molecules):
            inds[i] = mol.match_to_ref(ref_bundle, **kwargs)
        return inds


def import_molecule(fname, fmt='auto', **kwargs):
    """Imports geometry in provided format to Molecule object.

    Parameters
    ----------
    fname : file or str
        The open input file or filename.
    fmt : str, optional
        The file format. Default is auto (i.e. :func:`fileio.read_auto`).
    kwargs : dict, optional
        Additional keyword arguments for the read function.

    Returns
    -------
    Molecule
        The imported Molecule object.
    """
    moldat = fileio.read_multiple(fname, fmt=fmt, **kwargs)
    return Molecule(*moldat[0])


def import_bundle(fnamelist, fmt='auto', **kwargs):
    """Imports geometries in provided format to MoleculeBundle object.

    The fnamelist keyword can be a single filename or a list of
    filenames. If fmt='auto', different files may have different formats.

    Parameters
    ----------
    fnamelist : array_like
        The open input files or filenames.
    fmt : str, optional
        The file format. Default is auto (i.e. :func:`fileio.read_auto`).
    kwargs : dict, optional
        Additional keyword arguments for the read function.

    Returns
    -------
    MoleculeBundle
        The imported MoleculeBundle object.
    """
    moldat = fileio.read_multiple(fnamelist, fmt=fmt, **kwargs)
    molecules = [Molecule(*dat) for dat in moldat]
    return MoleculeBundle(molecules)


def _rearrange_inds(new_ind, old_ind):
    """Checks indices of rearrangement routines and returns indices in
    a single format.

    Parameters
    ----------
    new_ind : array_like
            New indices for the elements.
    old_ind : array_like
            Old indices for the elements.

    Returns
    -------
    new : ndarray
        An array of the new ordering of both old an new indices.
    old : ndarray
        An array of the old ordering of both old an new indices.

    Raises
    ------
    IndexError
        When the shape of new_ind and old_ind don't match.
    """
    new = np.atleast_1d(new_ind)
    old = np.atleast_1d(old_ind)
    if len(new) != len(old):
        raise IndexError('Old and new indices must be the same length')
    return np.hstack((old, new)), np.hstack((new, old))


def _add_type(inp1, inp2):
    """Checks the type of an object being added to a bundle.

    Parameters
    ----------
    inp1 : Molecule or MoleculeBundle
        The first input for concatenation into a list of molecules.
    inp2 : Molecule or MoleculeBundle
        The second input for concatenation into a list of molecules.

    Returns
    -------
    ndarray
        An array of Molecule objects from inp1 and inp2.

    Raises
    ------
    TypeError
        When the different types can't be added to make a MoleculeBundle
        object.
    """
    if isinstance(inp1, MoleculeBundle) and isinstance(inp2, MoleculeBundle):
        molecules = np.hstack((inp1.molecules, inp2.molecules))
    elif isinstance(inp1, MoleculeBundle) and isinstance(inp2, Molecule):
        molecules = np.hstack((inp1.molecules, inp2))
    elif isinstance(inp1, Molecule) and isinstance(inp2, MoleculeBundle):
        molecules = np.hstack((inp1, inp2.molecules))
    elif isinstance(inp1, Molecule) and isinstance(inp2, Molecule):
        molecules = np.hstack((inp1, inp2))
    else:
        fmt = 'Addition not supported for types \'{:s}\' and \'{:s}\'.'
        raise TypeError(fmt.format(type(inp1).__class__.__name__,
                                   type(inp2).__class__.__name__))
    return molecules



"""
Tests for the measure module.
"""
import pytest
import numpy as np
import gimbal.measure as measure
from examples import Geometries as eg


def test_stre_default():
    stre = measure.stre(eg.c2h4_ms[1], 0, 1)
    assert np.isclose(stre, 1.33496)


def test_stre_bohr():
    stre = measure.stre(eg.c2h4_ms[1], 0, 1, units='bohr')
    assert np.isclose(stre, 2.52270879)


def test_bend_default():
    bend = measure.bend(eg.c2h4_ms[1], 4, 1, 5)
    assert np.isclose(bend, 2.03464239)


def test_bend_degrees():
    bend = measure.bend(eg.c2h4_ms[1], 4, 1, 5, units='deg')
    assert np.isclose(bend, 116.57642205)


def test_tors_default():
    tors = measure.tors(eg.c2h4_ms[1], 2, 0, 1, 5)
    assert np.isclose(tors, -3.01462393)


def test_tors_degrees():
    tors = measure.tors(eg.c2h4_ms[1], 2, 0, 1, 5, units='deg')
    assert np.isclose(tors, -172.72522822)


def test_tors_absv():
    tors = measure.tors(eg.c2h4_ms[1], 2, 0, 1, 5, absv=True)
    assert np.isclose(tors, 3.01462393)


def test_oop_default():
    oop = measure.oop(eg.c2h4_ms[1], 1, 3, 2, 0)
    assert np.isclose(oop, -np.pi/4)


def test_oop_greater_than_pi():
    xyz = np.array([[    0.,             0.,             0.],
                    [-1./2.,             0., -np.sqrt(3)/2.],
                    [-1./2.,  np.sqrt(3)/2.,             0.],
                    [-1./2., -np.sqrt(3)/2.,             0.]])
    oop = measure.oop(xyz, 1, 2, 3, 0)
    assert np.isclose(oop, -2*np.pi/3)


def test_oop_degrees():
    oop = measure.oop(eg.c2h4_ms[1], 1, 3, 2, 0, units='deg')
    assert np.isclose(oop, -45.)


def test_oop_absv():
    oop = measure.oop(eg.c2h4_ms[1], 1, 3, 2, 0, absv=True)
    assert np.isclose(oop, np.pi/4)


def test_planeang_default():
    planeang = measure.planeang(eg.c2h4_ms[1], 2, 3, 0, 1, 4, 5)
    assert np.isclose(planeang, -1.93216345)


def test_planeang_degrees():
    planeang = measure.planeang(eg.c2h4_ms[1], 2, 3, 0, 1, 4, 5, units='deg')
    assert np.isclose(planeang, -110.70481105)


def test_planeang_absv():
    planeang = measure.planeang(eg.c2h4_ms[1], 2, 3, 0, 1, 4, 5, absv=True)
    assert np.isclose(planeang, 1.93216345)


def test_planetors_default():
    planetors = measure.planetors(eg.c2h4_ms[1], 2, 3, 0, 1, 4, 5)
    assert np.isclose(planetors, -2*np.pi/3)


def test_planetors_degrees():
    planetors = measure.planetors(eg.c2h4_ms[1], 2, 3, 0, 1, 4, 5, units='deg')
    assert np.isclose(planetors, -120.)


def test_planetors_absv():
    planetors = measure.planetors(eg.c2h4_ms[1], 2, 3, 0, 1, 4, 5, absv=True)
    assert np.isclose(planetors, 2*np.pi/3)


def test_edgetors_default():
    edgetors = measure.edgetors(eg.c2h4_ms[1], 2, 3, 0, 1, 4, 5)
    assert np.isclose(edgetors, -2.48199179)


def test_edgetors_degrees():
    edgetors = measure.edgetors(eg.c2h4_ms[1], 2, 3, 0, 1, 4, 5, units='deg')
    assert np.isclose(edgetors, -142.20765430)


def test_edgetors_absv():
    edgetors = measure.edgetors(eg.c2h4_ms[1], 2, 3, 0, 1, 4, 5, absv=True)
    assert np.isclose(edgetors, 2.48199179)



"""
Routines for measuring internal coordinate values from a cartesian
geometry.
"""
import numpy as np
import gimbal.constants as con


def stre(xyz, *inds, units='ang'):
    """Returns bond length based on index.

    Parameters
    ----------
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.
    inds : list
        The indices for which the bond distance is measured.
    units : str, optional
        The units of length for the output. Default is Angstroms.

    Returns
    -------
    float
        The bond length.
    """
    coord = np.linalg.norm(xyz[inds[0]] - xyz[inds[1]])
    return coord * con.conv('ang', units)


def bend(xyz, *inds, units='rad'):
    """Returns bending angle for 3 atoms in a chain based on index.

    Parameters
    ----------
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.
    inds : list
        The indices for which the bond angle is measured.
    units : str, optional
        The units of angle for the output. Default is radians.

    Returns
    -------
    float
        The bond angle.
    """
    e1 = con.unit_vec(xyz[inds[0]] - xyz[inds[1]])
    e2 = con.unit_vec(xyz[inds[2]] - xyz[inds[1]])

    coord = con.arccos(np.dot(e1, e2))
    return coord * con.conv('rad', units)


def tors(xyz, *inds, units='rad', absv=False):
    """Returns dihedral angle for 4 atoms in a chain based on index.

    Parameters
    ----------
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.
    inds : list
        The indices for which the dihedral angle is measured.
    units : str, optional
        The units of angle for the output. Default is radians.
    absv : bool, optional
        Specifies if the absolute value is returned.

    Returns
    -------
    float
        The dihedral angle.
    """
    e1 = xyz[inds[0]] - xyz[inds[1]]
    e2 = xyz[inds[2]] - xyz[inds[1]]
    e3 = xyz[inds[2]] - xyz[inds[3]]

    # get normals to 3-atom planes
    cp1 = con.unit_vec(np.cross(e1, e2))
    cp2 = con.unit_vec(np.cross(e2, e3))

    if absv:
        coord = con.arccos(np.dot(cp1, cp2))
    else:
        # get cross product of plane normals for signed dihedral angle
        cp3 = np.cross(cp1, cp2)
        coord = np.sign(np.dot(cp3, e2)) * con.arccos(np.dot(cp1, cp2))

    return coord * con.conv('rad', units)


def oop(xyz, *inds, units='rad', absv=False):
    """Returns out-of-plane angle of atom 1 connected to atom 4 in the
    2-3-4 plane.

    Contains an additional sign convention such that rotation of the
    out-of-plane atom over (under) the central plane atom gives an angle
    greater than :math:`\pi/2` (less than :math:`-\pi/2`).

    Parameters
    ----------
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.
    inds : list
        The indices for which the out-of-plane angle is measured.
    units : str, optional
        The units of angle for the output. Default is radians.
    absv : bool, optional
        Specifies if the absolute value is returned.

    Returns
    -------
    float
        The out-of-plane angle.
    """
    e1 = con.unit_vec(xyz[inds[0]] - xyz[inds[3]])
    e2 = con.unit_vec(xyz[inds[1]] - xyz[inds[3]])
    e3 = con.unit_vec(xyz[inds[2]] - xyz[inds[3]])

    sintau = np.dot(np.cross(e2, e3) / np.sqrt(1 - np.dot(e2, e3) ** 2), e1)
    coord = np.sign(np.dot(e2+e3, e1)) * con.arccos(sintau) + np.pi/2
    # sign convention to keep |oop| < pi
    if coord > np.pi:
        coord -= 2 * np.pi
    if absv:
        return abs(coord) * con.conv('rad', units)
    else:
        return coord * con.conv('rad', units)


def planeang(xyz, *inds, units='rad', absv=False):
    """Returns the angle between the 1-2-3 and 4-5-6 planes.

    Parameters
    ----------
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.
    inds : list
        The indices for which the plane angle is measured.
    units : str, optional
        The units of angle for the output. Default is radians.
    absv : bool, optional
        Specifies if the absolute value is returned.

    Returns
    -------
    float
        The plane angle.
    """
    e1 = xyz[inds[0]] - xyz[inds[2]]
    e2 = xyz[inds[1]] - xyz[inds[2]]
    e3 = xyz[inds[3]] - xyz[inds[2]]
    e4 = xyz[inds[4]] - xyz[inds[3]]
    e5 = xyz[inds[5]] - xyz[inds[3]]

    # get normals to 3-atom planes
    cp1 = con.unit_vec(np.cross(e1, e2))
    cp2 = con.unit_vec(np.cross(e4, e5))

    if absv:
        coord = con.arccos(np.dot(cp1, cp2))
    else:
        # get cross product of plane norms for signed dihedral angle
        cp3 = np.cross(cp1, cp2)
        coord = np.sign(np.dot(cp3, e3)) * con.arccos(np.dot(cp1, cp2))

    return coord * con.conv('rad', units)


def planetors(xyz, *inds, units='rad', absv=False):
    """Returns the plane angle with the central bond projected out.

    Parameters
    ----------
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.
    inds : list
        The indices for which the plane dihedral angle is measured.
    units : str, optional
        The units of angle for the output. Default is radians.
    absv : bool, optional
        Specifies if the absolute value is returned.

    Returns
    -------
    float
        The plane dihedral angle.
    """
    e1 = xyz[inds[0]] - xyz[inds[2]]
    e2 = xyz[inds[1]] - xyz[inds[2]]
    e3 = con.unit_vec(xyz[inds[3]] - xyz[inds[2]])
    e4 = xyz[inds[4]] - xyz[inds[3]]
    e5 = xyz[inds[5]] - xyz[inds[3]]

    # get normals to 3-atom planes
    cp1 = np.cross(e1, e2)
    cp2 = np.cross(e4, e5)

    # project out component along central bond
    pj1 = con.unit_vec(cp1 - np.dot(cp1, e3) * e3)
    pj2 = con.unit_vec(cp2 - np.dot(cp2, e3) * e3)

    if absv:
        coord = con.arccos(np.dot(pj1, pj2))
    else:
        # get cross product of vectors for signed dihedral angle
        cp3 = np.cross(pj1, pj2)
        coord = np.sign(np.dot(cp3, e3)) * con.arccos(np.dot(pj1, pj2))

    return coord * con.conv('rad', units)


def edgetors(xyz, *inds, units='rad', absv=False):
    """Returns the torsional angle based on the vector difference of the
    two external atoms (1-2 and 5-6) to the central 3-4 bond.

    Parameters
    ----------
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.
    inds : list
        The indices for which the edge dihedral angle is measured.
    units : str, optional
        The units of angle for the output. Default is radians.
    absv : bool, optional
        Specifies if the absolute value is returned.

    Returns
    -------
    float
        The edge dihedral angle.
    """
    e1 = con.unit_vec(xyz[inds[0]] - xyz[inds[2]])
    e2 = con.unit_vec(xyz[inds[1]] - xyz[inds[2]])
    e3 = con.unit_vec(xyz[inds[3]] - xyz[inds[2]])
    e4 = con.unit_vec(xyz[inds[4]] - xyz[inds[3]])
    e5 = con.unit_vec(xyz[inds[5]] - xyz[inds[3]])

    # take the difference between unit vectors of external bonds
    e2 -= e1
    e5 -= e4

    # get cross products of difference vectors and the central bond
    cp1 = con.unit_vec(np.cross(e2, e3))
    cp2 = con.unit_vec(np.cross(e3, e5))

    if absv:
        coord = con.arccos(np.dot(cp1, cp2))
    else:
        # get cross product of vectors for signed dihedral angle
        cp3 = np.cross(cp1, cp2)
        coord = np.sign(np.dot(cp3, e3)) * con.arccos(np.dot(cp1, cp2))

    return coord * con.conv('rad', units)
